/******************************************************************************
UTPlus
Copyright (c) 2005-2006 by Wormbo

This program is free software; you can redistribute and/or modify 
it under the terms of the Open Unreal Mod License version 1.1.
http://wiki.beyondunreal.com/wiki/OpenUnrealModLicense


Builds a network of camera paths.
******************************************************************************/


class UTP_CamPathManager extends Info;


//=============================================================================
// Consts
//=============================================================================

const INFINITY        = 999999;
const NO_CONNECTION   = -1;


//=============================================================================
// Structs
//=============================================================================

struct TPointDistance {
  var() int Index;
  var() float Distance;
  var() bool bVisible;
};

//=============================================================================
// Variables
//=============================================================================

var() transient UTP_MapPathNetwork PathData;
var() transient array<Teleporter> Teleporters;
var() transient array<WarpZoneMarker> WarpZones;
var() transient array<int> TeleporterIndices, WarpZoneIndices;
var transient int IterationCounter, CurrNavPoint;
var transient bool bDirty;

var transient vector LastPath, LastPathSource, LastPathDest;
var transient float LastPathTime;


//== static GetPathManager ====================================================
/**
Returns an existing UTP_CamPathManager or spawns and returns a new one.
*/
//=============================================================================

static function UTP_CamPathManager GetPathManager(Actor Referencer)
{
  local UTP_CamPathManager PathManager;
  
  if ( Referencer == None )
    return None;
  
  foreach Referencer.DynamicActors(class'UTP_CamPathManager', PathManager) {
    return PathManager;
  }
  
  return Referencer.Spawn(class'UTP_CamPathManager', None);
}


//== DrawStatus =============================================================
/**
Stub function for drawing the status of the path calculations.
*/
//===========================================================================

function DrawStatus(Canvas C);


//== DrawStatus =============================================================
/**
Stub function for drawing the status of the path calculations.
*/
//===========================================================================

event Destroyed()
{
  SavePathNetwork();
  PathData = None;
}


//== SavePathNetwork ========================================================
/**
Attempts to save the path network to disk.
*/
//===========================================================================

function SavePathNetwork()
{
  local GameInfo Game;
  
  if ( !bDirty )
    return;
  
  Game = Level.Game;
  if ( Game == None )
    Game = Level.GetLocalPlayerController().GetEntryLevel().Game;
  
  if ( Game == None || !Game.SavePackage("UTPPaths_" $ Outer) )
    log("Error: Failed to save path data. Paths will need to be rebuild after reloading this map!", 'UTPlus');
  else
    log("Path data saved to UTPPaths_" $ Outer, 'UTPlus');
  
  bDirty = False;
  SetTimer(30.0, False);
}


//== state GenerateNetwork ====================================================
/**
Generates the path network.
*/
//=============================================================================

auto state GenerateNetwork
{
  //== FindPath ===============================================================
  /**
  Paths are not yet known, so we can't use the path data.
  */
  //===========================================================================
  
  function bool FindPath(vector Source, vector Destination, out vector Path)
  {
    Path = Destination;
    return false;
  }
  
  
  //== IsInValidZone ==========================================================
  /**
  Returns, whether the actor is in a valid zone.
  */
  //===========================================================================
  
  function bool IsInValidZone(Actor Other)
  {
    local vector HL, HN;
    
    if ( Other.Region.ZoneNumber == 0 || SkyZoneInfo(Other.Region.Zone) != None )
      return false; // outside level geometry or in skyzone
    
    return Trace(HL, HN, Other.Location - vect(15,15,15), Other.Location + vect(15,15,15), False, vect(22,22,22)) == None;
  }
  
  
  //== FindNodes ==============================================================
  /**
  Fills the node, teleporter and warp zone lists.
  */
  //===========================================================================
  
  function FindNodes()
  {
    local NavigationPoint NavPoint;
    local KeyPoint KeyPoint;
    local ZoneInfo Zone;
    local Light Light;
    local int i;
    //local vector BaseLocation;
    local GameInfo Game;
    
    if (PathData == None)
      PathData = UTP_MapPathNetwork(FindObject("UTPPaths_" $ Outer $ ".MapPathData", class'UTP_MapPathNetwork'));
    
    if (PathData == None) {
      Game = Level.Game;
      if ( Game == None )
        Game = Level.GetLocalPlayerController().GetEntryLevel().Game;
      
      if ( Game != None ) {
        PathData = Game.LoadDataObject(class'UTP_MapPathNetwork', "MapPathData", "UTPPaths_" $ Outer);
        if ( PathData == None ) {
          PathData = Game.CreateDataObject(class'UTP_MapPathNetwork', "MapPathData", "UTPPaths_" $ Outer);
          if ( PathData == None )
            Error("Could not create path data object!");
        }
        else if ( PathData.Node.Length > 0 ) {
          log("Path data loaded from UTPPaths_" $ Outer, 'UTPlus');
          GotoState('CalculateShortestPaths');
          return;
        }
      }
      else
        Error("No GameInfo available!");
    }
    
    foreach AllActors(class'ZoneInfo', Zone) {
      if ( !IsInValidZone(Zone) )
        continue;
      PathData.Node.Length = i + 1;
      //Markers[i]                    = Zone;
      //PathData.Node[i].MarkerName   = string(KeyPoint.Name);
      //PathData.Node[i].MarkerClass  = KeyPoint.Class;
      PathData.Node[i].Location     = Zone.Location;
      ++i;
    }
    foreach AllActors(class'KeyPoint', KeyPoint) {
      if ( !IsInValidZone(KeyPoint) )
        continue;
      PathData.Node.Length = i + 1;
      //Markers[i]                    = KeyPoint;
      //PathData.Node[i].MarkerName   = string(KeyPoint.Name);
      //PathData.Node[i].MarkerClass  = KeyPoint.Class;
      PathData.Node[i].Location     = KeyPoint.Location;
      ++i;
    }
    for (NavPoint = Level.NavigationPointList; NavPoint != None; NavPoint = NavPoint.NextNavigationPoint) {
      //RealLocation = NavPoint.Location;
      if ( !IsInValidZone(NavPoint) )
        continue;
      /*if ( LiftCenter(NavPoint) != None && Mover(NavPoint.Base) != None && Mover(NavPoint.Base).NumKeys > 1 ) {
        // add one path point for each key frame of the LiftCenter's base mover
        BaseLocation = NavPoint.Location - Mover(NavPoint.Base).KeyPos[Mover(NavPoint.Base).KeyNum];
        for (j = 0; j < Mover(NavPoint.Base).NumKeys; ++j) {
          PathData.Node.Length = i + 1;
          PathData.Node[i++].Location = BaseLocation + Mover(NavPoint.Base).KeyPos[j];
        }
      }
      else {*/
        PathData.Node.Length = i + 1;
        //Markers[i]                    = NavPoint;
        //PathData.Node[i].MarkerName   = string(NavPoint.Name);
        //PathData.Node[i].MarkerClass  = NavPoint.Class;
        PathData.Node[i].Location     = NavPoint.Location;
        if ( Teleporter(NavPoint) != None ) {
          TeleporterIndices[TeleporterIndices.Length] = i;
          Teleporters[Teleporters.Length] = Teleporter(NavPoint);
        }
        else if ( WarpZoneMarker(NavPoint) != None && WarpZoneMarker(NavPoint).markedWarpZone != None ) {
          WarpZoneIndices[WarpZoneIndices.Length] = i;
          WarpZones[WarpZones.Length] = WarpZoneMarker(NavPoint);
        }
        ++i;
      //}
    }
    foreach AllActors(class'Light', Light) {
      if ( !IsInValidZone(Light) )
        continue;
      PathData.Node.Length = i + 1;
      //Markers[i]                    = Light;
      //PathData.Node[i].MarkerName   = string(KeyPoint.Name);
      //PathData.Node[i].MarkerClass  = KeyPoint.Class;
      PathData.Node[i].Location     = Light.Location;
      ++i;
    }
    
    // initialize the additional navpoint variables
    for (i = 0; i < PathData.Node.Length; ++i) {
      PathData.Node[i].DistanceTo.Length   = PathData.Node.Length;
      PathData.Node[i].BestPathFrom.Length = PathData.Node.Length;
      PathData.Node[i].BestPathFrom[i]     = i; // the node is always the best path to itself
    }
  }
  
  
  //== FindConnections ========================================================
  /**
  Finds all direct connnections to other navigations points for a single point.
  */
  //===========================================================================
  
  function FindConnections(int NavPointIndex)
  {
    local int i;
    local vector HL, HN;
    
    //SetCollision(True, True, True);
    for (i = NavPointIndex + 1; i < PathData.Node.Length; ++i) {
      if ( VSize(PathData.Node[i].Location - PathData.Node[NavPointIndex].Location) < 10000 && FastTrace(PathData.Node[i].Location, PathData.Node[NavPointIndex].Location) && Trace(HL, HN, PathData.Node[i].Location, PathData.Node[NavPointIndex].Location, False, vect(22,22,22)) == None ) {
        // found visible navigation point
        //PathData.Node[NavPointIndex].Successors[PathData.Node[NavPointIndex].Successors.Length] = i;
        PathData.Node[NavPointIndex].DistanceTo[i] = VSize(PathData.Node[i].Location - PathData.Node[NavPointIndex].Location);
        PathData.Node[NavPointIndex].BestPathFrom[i] = i;
        //PathData.Node[i].Successors[PathData.Node[i].Successors.Length] = NavPointIndex;
        PathData.Node[i].DistanceTo[NavPointIndex] = PathData.Node[NavPointIndex].DistanceTo[i];
        PathData.Node[i].BestPathFrom[NavPointIndex] = NavPointIndex;
      }
      else {
        // no LOS between the two navigation points
        PathData.Node[NavPointIndex].DistanceTo[i]   = INFINITY;
        PathData.Node[i].DistanceTo[NavPointIndex]   = INFINITY;
        PathData.Node[NavPointIndex].BestPathFrom[i] = NO_CONNECTION;
        PathData.Node[i].BestPathFrom[NavPointIndex] = NO_CONNECTION;
      }
      if ( PathData.MaxDistance < VSize(PathData.Node[i].Location - PathData.Node[NavPointIndex].Location) )
        PathData.MaxDistance = VSize(PathData.Node[i].Location - PathData.Node[NavPointIndex].Location);
    }
  }
  
  
  //== ConnectTeleporters =====================================================
  /**
  Adds direct connections for matching teleporters.
  */
  //===========================================================================
  
  function ConnectTeleporters()
  {
    local int i, j;
    
    for (i = 0; i < Teleporters.Length; ++i) {
      if ( Teleporters[i].bEnabled && Teleporters[i].URL != "" ) {
        for (j = 0; j < Teleporters.Length; ++j) {
          if ( i != j && Teleporters[i].URL ~= string(Teleporters[j].Tag) ) {
            PathData.Node[TeleporterIndices[i]].DistanceTo[TeleporterIndices[j]]   = 0;
            PathData.Node[TeleporterIndices[j]].BestPathFrom[TeleporterIndices[i]] = TeleporterIndices[i];
          }
        }
      }
    }
    Teleporters.Length = 0;
    TeleporterIndices.Length = 0;
  }
  
  
  //== ConnectTeleporters =====================================================
  /**
  Adds direct connections for matching warp zone markers.
  */
  //===========================================================================
  
  function ConnectWarpZones()
  {
    local int i, j;
    
    for (i = 0; i < WarpZones.Length; ++i) {
      for (j = 0; j < WarpZones.Length; ++j) {
        if ( WarpZones[i].markedWarpZone.OtherSideActor == WarpZones[j].markedWarpZone ) {
          PathData.Node[WarpZoneIndices[i]].DistanceTo[WarpZoneIndices[j]]   = 0;
          PathData.Node[WarpZoneIndices[j]].BestPathFrom[WarpZoneIndices[i]] = WarpZoneIndices[i];
        }
      }
    }
    WarpZones.Length = 0;
    WarpZoneIndices.Length = 0;
  }
  
  
  //== DrawStatus ===========================================================
  /**
  Stub function for drawing the status of the path calculations.
  */
  //=========================================================================
  
  function DrawStatus(Canvas C)
  {
    C.Reset();
    C.SetDrawColor(255, 255, 255);
    if ( PathData == None ) {
      C.DrawScreenText("Initializing camera path calculations...", 0.5, 0.999, DP_LowerMiddle);
    }
    else if ( CurrNavPoint < PathData.Node.Length ) {
      C.DrawScreenText("Connecting path nodes: ", 0.5, 0.999, DP_LowerRight);
      C.DrawScreenText(100.0 * float(CurrNavPoint) / float(PathData.Node.Length) $ "%", 0.5, 0.999, DP_LowerLeft);
    }
    else {
      C.DrawScreenText("Completing path node connections...", 0.5, 0.999, DP_LowerMiddle);
    }
  }
  
Begin:
  Sleep(0.0);
  FindNodes();
  log("Found" @ PathData.Node.Length @ "path points", 'UTPlus');
  while (CurrNavPoint < PathData.Node.Length) {
    if ( CurrNavPoint % Max(1000 / PathData.Node.Length, 1) == 0 )
      Sleep(0.0);
    //log("Finding connections for point" @ CurrNavPoint @ "-" @ PathData.Node[CurrNavPoint].Location);
    FindConnections(CurrNavPoint);
    ++CurrNavPoint;
  }
  if ( Teleporters.Length > 0 ) {
    Sleep(0.0);
    log("Connecting Teleporters", 'UTPlus');
    ConnectTeleporters();
  }
  if ( WarpZones.Length > 0 ) {
    Sleep(0.0);
    log("Connecting WarpZones", 'UTPlus');
    ConnectWarpZones();
  }
  //Markers.Length = 0;
  bDirty = True;  // don't try to save before this point
  log("Path network complete.", 'UTPlus');
  Sleep(0.0);
  GotoState('CalculateShortestPaths');
}


//== state CalculateShortestPaths =============================================
/**
Calculates the shortest paths between each pair of path points in the network.
*/
//=============================================================================

state CalculateShortestPaths
{
  //== DrawStatus ===========================================================
  /**
  Stub function for drawing the status of the path calculations.
  */
  //=========================================================================
  
  function DrawStatus(Canvas C)
  {
    local float ProgressCount, ProgressTotal;
    
    ProgressCount = PathData.k + PathData.Node.Length * (PathData.i + PathData.Node.Length * PathData.j);
    ProgressTotal = PathData.Node.Length ** 3;
    
    C.Reset();
    C.SetDrawColor(255, 255, 255);
    C.DrawScreenText("Calculating camera paths: ", 0.5, 0.999, DP_LowerRight);
    C.DrawScreenText(" " $ ProgressCount @ "/" @ ProgressTotal @ "(" $ 100.0 * ProgressCount / ProgressTotal $ "%)", 0.5, 0.999, DP_LowerLeft);
  }
  
  function Reset()
  {
    SavePathNetwork();
  }
  
  function NotifyLocalPlayerDead(PlayerController PC)
  {
    SetTimer(0.1, False);
  }
  
  event Timer()
  {
    if ( Level.GetLocalPlayerController().Pawn == None )
      SavePathNetwork();
    else
      SetTimer(60.0, false);
  }
  
Begin:
  Sleep(0.0);
  Tag = 'NewPath';
  if ( !PathData.bComplete ) {
    Timer();
    log("Calculating shortest paths...", 'UTPlus');
    // calculate all shortest paths with Floyd's algorithm
    while (PathData.j < PathData.Node.Length) {
      while (PathData.i < PathData.Node.Length) {
        if ( PathData.Node[PathData.i].DistanceTo[PathData.j] < INFINITY ) {
          if ( !Level.GetLocalPlayerController().bDemoOwner && (!class'UTP_ClientSettings'.default.bCalcPathsDuringGame && !Level.GetLocalPlayerController().PlayerReplicationInfo.bOnlySpectator || Level.GetLocalPlayerController().Pawn != None) ) {
            //SavePathNetwork();
            Sleep(0.0);
            continue;
          }
          while (PathData.k < PathData.Node.Length) {
            if ( IterationCounter > 4000 ) {
              IterationCounter -= 4000;
              //log("Iteration" @ PathData.j * Square(PathData.Node.Length) + PathData.i * PathData.Node.Length + PathData.k @ "of" @ PathData.Node.Length ** 3);
              Sleep(0.0);
            }
            if ( PathData.Node[PathData.i].DistanceTo[PathData.k] > PathData.Node[PathData.i].DistanceTo[PathData.j] +  PathData.Node[PathData.j].DistanceTo[PathData.k] ) {
              // found a path shorter than the currently known path from node i to node k
              PathData.Node[PathData.i].DistanceTo[PathData.k] = PathData.Node[PathData.i].DistanceTo[PathData.j] +  PathData.Node[PathData.j].DistanceTo[PathData.k];
              
              PathData.Node[PathData.k].BestPathFrom[PathData.i] = PathData.Node[PathData.k].BestPathFrom[PathData.j];
              
              IterationCounter += 5;
            }
            else
              IterationCounter += 3;
            
            ++PathData.k;
          } // while k
          
          // prepare for next loop
          PathData.k = 0;
        }
        
        if ( !bDirty )
          bDirty = True;
        ++PathData.i;
      } // while i
      
      // prepare for next loop
      PathData.i = 0;
      
      ++PathData.j;
    } // while j
    
    // set to 0 just for neatness, we won't need it anymore :)
    PathData.j = 0;
    
    log("Shortest paths calculations complete.", 'UTPlus');
    PathData.bComplete = True;
    GotoState('WaitForPackageSaved');
  }
  GotoState('');
}

//== state WaitForPackageSaved ================================================
/**
Waits for a good opportunity to save the path data package.
*/
//=============================================================================

state WaitForPackageSaved extends CalculateShortestPaths
{
  //== DrawStatus ===========================================================
  /**
  Stub function for drawing the status of the path calculations.
  */
  //=========================================================================
  
  function DrawStatus(Canvas C)
  {
    C.Reset();
    C.SetDrawColor(255, 255, 255);
    C.DrawScreenText("Waiting to save path data...", 0.5, 0.999, DP_LowerMiddle);
  }
  
  function SavePathNetwork()
  {
    Global.SavePathNetwork();
    GotoState('');
  }
  
Begin:
}


//== GetPath ==================================================================
/**
Finds the shortest path from the starting node to the goal node and returns it in Path. Returns, whether a path is (already) known.
*/
//=============================================================================

function bool GetPath(int StartNode, int GoalNode, out array<vector> Path)
{
  Path.Length = 0;
  
  if ( PathData.Node[GoalNode].BestPathFrom[StartNode] == NO_CONNECTION ) {
    return false;
  }
  while(StartNode != GoalNode) {
    Path.Insert(0, 1);
    Path[0] = PathData.Node[GoalNode].Location;
    GoalNode = PathData.Node[GoalNode].BestPathFrom[StartNode];
  }
  Path.Insert(0, 1);
  Path[0] = PathData.Node[StartNode].Location;
  
  return Path.Length > 0;
}


//== FindNodesByDistance ======================================================
/**
Returns an array of all nodes, sorted by distance to the specified reference location.
*/
//=============================================================================

function FindNodesByDistance(vector RefLocation, out array<TPointDistance> Nodes, float MaxDist, optional float MinDist)
{
  local int i;
  local float Dist;
  local int InsertIndex, High, Closest;
  
  if ( PathData == None ) {
    Nodes.Length = 0;
    return;
  }
  
  // preallocate memory
  Nodes.Length = PathData.Node.Length;
  Nodes.Length = 0;
  
  for (i = 0; i < PathData.Node.Length; ++i) {
    Dist = VSize(PathData.Node[i].Location - RefLocation);
    if ( Dist > MaxDist || Dist < MinDist )
      continue;
    
    // find the insert location with binary search
    InsertIndex = 0;
    High = Nodes.Length - 1;
    while (InsertIndex <= High) {
      Closest = (InsertIndex + High) / 2;
      if ( Nodes[Closest].Distance == Dist ) {
        InsertIndex = Closest;
        break;
      }
      if ( Nodes[Closest].Distance > Dist )
        High = Closest - 1;
      else if ( Nodes[Closest].Distance < Dist )
        InsertIndex = Closest + 1;
    }
    if ( Nodes.Length != 0 && InsertIndex < Nodes.Length && Nodes[InsertIndex].Distance < Dist )
      ++InsertIndex;
    
    // insert at this location
    Nodes.Insert(InsertIndex, 1);
    Nodes[InsertIndex].Index    = i;
    Nodes[InsertIndex].Distance = Dist;
  }
}


//== FindPath =================================================================
/**
Returns the next path vector on the shortest path from source to destination.
*/
//=============================================================================

function bool FindPath(vector Source, vector Destination, out vector Path)
{
  local int i, j, s, d;
  local int BestDest, BestSource, BestRoute;
  local float BestDist, PathDist;
  local array<TPointDistance> Sources, Destinations;
  local vector HL, HN;
  
  if ( PathData == None ) {
    //ClearStayingDebugLines();
    //DrawDebugLine(Source, Destination, 255, 0, 0);
    Path = Destination;
    return False;
  }
  
  if ( Trace(HL, HN, Destination, Source, False, vect(15,15,15)) == None ) {
    Path = Destination;
    return True;
  }
  
  if ( Level.TimeSeconds - LastPathTime < 0.05 && VSize(Source - LastPathSource) < 20 && VSize(Destination - LastPathDest) < 100 && Trace(HL, HN, LastPath, Source, False, vect(10,10,10)) == None ) {
    Path = LastPath;
    return true;
  }
  
  //StopWatch(False);
  PathDist = (VSize(Destination - Source) + PathData.MaxDistance + 1000) / 5;
  //log("Finding sources");
  FindNodesByDistance(Source, Sources, PathDist);
  if ( Sources.Length > 20 )
    Sources.Length = 20;
  //log("Finding destinations");
  FindNodesByDistance(Destination, Destinations, PathDist);
  if ( Destinations.Length > 20 )
    Destinations.Length = 20;
  
  //log("Checking path lengths");
  BestDist = INFINITY;
  for (i = 0; i < Sources.Length; ++i) {
    s = Sources[i].Index;
    for (j = 0; j < Destinations.Length; ++j) {
      d = Destinations[j].Index;
      PathDist = PathData.Node[s].DistanceTo[d];
      if ( BestDist > Sources[i].Distance + PathDist + Destinations[j].Distance ) {
        if ( !Sources[i].bVisible ) {
          if ( Trace(HL, HN, PathData.Node[s].Location, Source, False, vect(10,10,10)) == None )
            Sources[i].bVisible = True;
          else
            break;  // skip this source point entirely
        } // if source visible
        
        if ( !Destinations[j].bVisible ) {
          if ( FastTrace(Destination, PathData.Node[d].Location) )
            Destinations[j].bVisible = True;
          else {
            Destinations.Remove(j--, 1);  // remove this point from the list of potential destinations
            continue;                     // skip this destination point
          }
        } // if destination visible
        
        BestDist    = Sources[i].Distance + PathDist + Destinations[j].Distance;
        BestDest    = d;
        BestSource  = s;
        
      } // if better distance
    } // for destinations
  } // for sources
  //StopWatch(True);
  
  if ( BestDist < INFINITY ) {
    LastPathSource  = Source;
    LastPathDest    = Destination;
    BestRoute       = BestDest;
    //ClearStayingDebugLines();
    //DrawStayingDebugLine(PathData.Node[BestDest].Location, Destination, 0, 255, 128);
    while (BestRoute != -1 && BestRoute != BestSource) {
      if ( Trace(HL, HN, PathData.Node[BestRoute].Location, Source, False, vect(15,15,15)) == None )
        break;
      
      //DrawStayingDebugLine(PathData.Node[BestRoute].Location, PathData.Node[PathData.Node[BestRoute].BestPathFrom[BestSource]].Location, 0, 160, 255);
      BestRoute = PathData.Node[BestRoute].BestPathFrom[BestSource];
    }
    //DrawStayingDebugLine(Source, PathData.Node[BestSource].Location, 128, 0, 255);
    //DrawStayingDebugLine(Source, PathData.Node[BestRoute].Location, 0, 0, 255);
    
    LastPath        = PathData.Node[BestRoute].Location;
    LastPathTime    = Level.TimeSeconds;
    Path = LastPath;
    return true;
  }
  //ClearStayingDebugLines();
  //DrawStayingDebugLine(Source, Destination, 255, 255, 255);
  Path = Destination;
  return false;
}


//== Trigger ==================================================================
/**
For testing: the console command "CauseEvent NewPath" will generate a random path and draw it. Use "ClearAllDebugLines" to remove the path lines.
*/
//=============================================================================

event Trigger(Actor Other, Pawn EventInstigator)
{
  local array<vector> Path;
  local int i;
  
  ClearStayingDebugLines();
  if ( PathData != None && GetPath(Rand(PathData.Node.Length), Rand(PathData.Node.Length), Path) ) {
    log("Drawing path");
    for (i = 1; i < Path.Length; ++i) {
      DrawStayingDebugLine(Path[i-1], Path[i], 0, 255, 0);
    }
  }
  else
    log("Path not available.");
}


//=============================================================================
// Default properties
//=============================================================================

defaultproperties
{
  bAlwaysTick=True
}