/******************************************************************************
UTPlus
Copyright (c) 2005-2006 by Wormbo

This program is free software; you can redistribute and/or modify 
it under the terms of the Open Unreal Mod License version 1.1.
http://wiki.beyondunreal.com/wiki/OpenUnrealModLicense


Client-side player GUID database.
******************************************************************************/


class UTP_PlayerIDsBase extends Object
    abstract;


//=============================================================================
// Structs
//=============================================================================

struct TPlayerIndexInfo {
  var array<int> iNames;  // indices of TNamesInfo array
  var int        iIP;     // index of TIPsInfo array
  var int        iID;     // index of TIDsInfo array
};

struct TNamesInfo {
  var string     PlayerName;
  var array<int> iPlayers;
};

struct TIPsInfo {
  var int        IP;
  var array<int> iPlayers;
};

struct TIDsInfo {
  var int        ID;
  var array<int> iPlayers;
};

struct THashTableEntry {
  var array<int> iIDs;
  var array<int> iIPs;
  var array<int> iNames;
};


//=============================================================================
// Variables
//=============================================================================

var protected array<TNamesInfo>  Names;
var protected array<TIPsInfo>    IPs;
var protected array<TIDsInfo>    IDs;

var protected array<TPlayerIndexInfo> PlayerList;

var protected editconst THashTableEntry HashTable[256];

var bool bIsDirty;


//== AddPlayer ================================================================
/**
Adds a new player to the database and returns the player's database ID.
*/
//=============================================================================

function int AddPlayer(int PlayerID, int PlayerIP, string PlayerName)
{
  local int iID, iIP, iPlayerName, PlayerIndex;
  local int i;
  
  if ( PlayerID == 0 || PlayerIP == 0 || PlayerName == "" )
    return -1;
  
  //log(PlayerID @ PlayerIP @ PlayerName, 'UTPlusDebug');
  iID = FindID(PlayerID, True);
  iIP = FindIP(PlayerIP, True);
  iPlayerName = FindName(PlayerName, True);
  
  for (i = 0; i < IPs[iIP].iPlayers.Length; ++i) {
    if ( PlayerList[IPs[iIP].iPlayers[i]].iID == iID ) {
      PlayerIndex = IPs[iIP].iPlayers[i];
      break;
    }
  }
  if ( PlayerIndex == 0 ) {
    PlayerIndex = Max(PlayerList.Length, 1);
    PlayerList.Length = PlayerIndex + 1;
    PlayerList[PlayerIndex].iID = iID;
    PlayerList[PlayerIndex].iIP = iIP;
    PlayerList[PlayerIndex].iNames[0] = iPlayerName;
    IDs[iID].iPlayers[IDs[iID].iPlayers.Length] = PlayerIndex;
    IPs[iIP].iPlayers[IPs[iIP].iPlayers.Length] = PlayerIndex;
    Names[iPlayerName].iPlayers[Names[iPlayerName].iPlayers.Length] = PlayerIndex;
    
    if ( !bIsDirty )
      bIsDirty = True;
  }
  else {
    AddPlayerNameIndex(PlayerIndex, iPlayerName);
  }
  
  return PlayerIndex;
}


//== AddPlayerName ============================================================
/**
Adds another player name for the specified database ID.
*/
//=============================================================================

function AddPlayerName(int PlayerIndex, string PlayerName)
{
  local int iPlayerName;
  
  //log("..." @ PlayerName);
  if ( PlayerIndex <= 0 || PlayerName == "" )
    return;
  
  iPlayerName = FindName(PlayerName, True);
  AddPlayerNameIndex(PlayerIndex, iPlayerName);
}


//== AddPlayerNameIndex =======================================================
/**
Adds the index of a player name for the specified database ID.
*/
//=============================================================================

protected function AddPlayerNameIndex(int PlayerIndex, int iPlayerName)
{
  local int i;
  
  for (i = 0; i < PlayerList[PlayerIndex].iNames.Length; ++i) {
    if ( PlayerList[PlayerIndex].iNames[i] == iPlayerName )
      return; // name index is already in the list 
  }
  
  if ( !bIsDirty )
    bIsDirty = True;
  
  // i is set to PlayerList[PlayerIndex].iNames.Length here
  PlayerList[PlayerIndex].iNames[i] = iPlayerName;
  
  // add the player to the list of players referencing this name
  for (i = 0; i < Names[iPlayerName].iPlayers.Length; ++i) {
    if ( Names[iPlayerName].iPlayers[i] == PlayerIndex )
      return; // player is already in the list
  }
  // i is set to Names[iPlayerName].iPlayers.Length here
  Names[iPlayerName].iPlayers[i] = PlayerIndex;
}


//== FindID ===================================================================
/**
Returns the index of an ID.
*/
//=============================================================================

function int FindID(int PlayerID, optional bool bAdd)
{
  local int InsertIndex, Closest;
  local byte HashIndex;
  local int Low, High;
  
  HashIndex = HashInt(PlayerID);
  High = HashTable[HashIndex].iIDs.Length - 1;
  
  while (Low <= High) {
    Closest = (Low + High) / 2;
    if ( IDs[HashTable[HashIndex].iIDs[Closest]].ID == PlayerID )
      return HashTable[HashIndex].iIDs[Closest];
    if ( IDs[HashTable[HashIndex].iIDs[Closest]].ID > PlayerID )
      High = Closest - 1;
    else if ( IDs[HashTable[HashIndex].iIDs[Closest]].ID < PlayerID )
      Low = Closest + 1;
  }
  
  if ( !bAdd )
    return -1;
  
  if ( HashTable[HashIndex].iIDs.Length != 0 && Low < HashTable[HashIndex].iIDs.Length && IDs[HashTable[HashIndex].iIDs[Low]].ID < PlayerID )
    Closest = Low + 1;
  else
    Closest = Low;
  
  // enter new ID into database
  InsertIndex = IDs.Length;
  IDs.Length = InsertIndex + 1;
  IDs[InsertIndex].ID = PlayerID;
  HashTable[HashIndex].iIDs.Insert(Closest, 1);
  HashTable[HashIndex].iIDs[Closest] = InsertIndex;
  
  if ( !bIsDirty )
    bIsDirty = True;
  
  return InsertIndex;
}


//== FindIP ===================================================================
/**
Returns the index of an IP.
*/
//=============================================================================

function int FindIP(int PlayerIP, optional bool bAdd)
{
  local int InsertIndex, Closest;
  local byte HashIndex;
  local int Low, High;
  
  HashIndex = HashInt(PlayerIP);
  High = HashTable[HashIndex].iIPs.Length - 1;
  
  while (Low <= High) {
    Closest = (Low + High) / 2;
    if ( IPs[HashTable[HashIndex].iIPs[Closest]].IP == PlayerIP )
      return HashTable[HashIndex].iIPs[Closest];
    if ( IPs[HashTable[HashIndex].iIPs[Closest]].IP > PlayerIP )
      High = Closest - 1;
    else if ( IPs[HashTable[HashIndex].iIPs[Closest]].IP < PlayerIP )
      Low = Closest + 1;
  }
  
  if ( !bAdd )
    return -1;
  
  if ( HashTable[HashIndex].iIPs.Length != 0 && Low < HashTable[HashIndex].iIPs.Length && IPs[HashTable[HashIndex].iIPs[Low]].IP < PlayerIP )
    Closest = Low + 1;
  else
    Closest = Low;
  
  // enter new ID into database
  InsertIndex = IPs.Length;
  IPs.Length = InsertIndex + 1;
  IPs[InsertIndex].IP = PlayerIP;
  HashTable[HashIndex].iIPs.Insert(Closest, 1);
  HashTable[HashIndex].iIPs[Closest] = InsertIndex;
  
  if ( !bIsDirty )
    bIsDirty = True;
  
  return InsertIndex;
}


//== FindName =================================================================
/**
Returns the index of a player name.
*/
//=============================================================================

function int FindName(string PlayerName, optional bool bAdd)
{
  local int InsertIndex, Closest;
  local byte HashIndex;
  local int Low, High;
  
  HashIndex = HashString(PlayerName);
  High = HashTable[HashIndex].iNames.Length - 1;
  
  while (Low <= High) {
    Closest = (Low + High) / 2;
    if ( Names[HashTable[HashIndex].iNames[Closest]].PlayerName == PlayerName )
      return HashTable[HashIndex].iNames[Closest];
    if ( Names[HashTable[HashIndex].iNames[Closest]].PlayerName > PlayerName )
      High = Closest - 1;
    else if ( Names[HashTable[HashIndex].iNames[Closest]].PlayerName < PlayerName )
      Low = Closest + 1;
  }
  
  if ( !bAdd )
    return -1;
  
  if ( HashTable[HashIndex].iNames.Length != 0 && Low < HashTable[HashIndex].iNames.Length && Names[HashTable[HashIndex].iNames[Low]].PlayerName < PlayerName )
    Closest = Low + 1;
  else
    Closest = Low;
  
  // enter new ID into database
  InsertIndex = Names.Length;
  Names.Length = InsertIndex + 1;
  Names[InsertIndex].PlayerName = PlayerName;
  HashTable[HashIndex].iNames.Insert(Closest, 1);
  HashTable[HashIndex].iNames[Closest] = InsertIndex;
  
  if ( !bIsDirty )
    bIsDirty = True;
  
  return InsertIndex;
}


//== GetAllPlayers ============================================================
/**
Return all player UIDs with the same ID.
*/
//=============================================================================

function array<int> GetAllPlayers(int iPlayerUID)
{
  if ( iPlayerUID < 0 || iPlayerUID >= PlayerList.Length ) {
    log("UTP_PlayerIDBase.GetAllPlayers - PlayerUID out of bounds:" @ iPlayerUID $ "/" $ PlayerList.Length, 'UTPlusError');
  }
  else
    return IDs[PlayerList[iPlayerUID].iID].iPlayers;
}


//== HashInt ==================================================================
/**
Calculates the hash value for an integer.
*/
//=============================================================================

final function byte HashInt(int i)
{
  return i ^ (i >>> 8) ^ (i >>> 16) ^ (i >>> 24);
}


//== HashString =================================================================
/**
Calculates the hash value for a string.
*/
//=============================================================================

final function byte HashString(string s)
{
  local int Hash;
  
  Hash = Len(s);
  Hash *= Asc(Mid(s, Hash / 2, 1));
  Hash *= Asc(s);
  Hash *= Asc(Right(s, 1));
  
  return byte(Hash);
}


//=============================================================================
// Default properties
//=============================================================================

defaultproperties
{
}
