/******************************************************************************
UTPlus
Copyright (c) 2005-2006 by Wormbo

This program is free software; you can redistribute and/or modify 
it under the terms of the Open Unreal Mod License version 1.1.
http://wiki.beyondunreal.com/wiki/OpenUnrealModLicense


Provides "Real-Time" timer functionality.
******************************************************************************/


class UTP_RealTime extends Info;


//=============================================================================
// Constants
//=============================================================================

const MILLISECONDS_PER_DAY = 86400000;


//=============================================================================
// Structs
//=============================================================================

struct TTimerInfo {
  var int TimerInterval;
  var int CurrentTime;
  var Actor Owner;
  var bool bRepeat;
  var bool bFixPrecisionError;
};


//=============================================================================
// Variables
//=============================================================================

var array<TTimerInfo> Timers;
var array<int> StopwatchStarts;

var int OldTime;


//== Tick =====================================================================
/**
Update the timers.
*/
//=============================================================================

function Tick(float dt)
{
  local int NewTime, DeltaTime;
  local int i;
  
  NewTime = Level.MilliSecond + 1000 * (Level.Second + 60 * (Level.Minute + 60 * Level.Hour));
  DeltaTime = NewTime - OldTime;
  if ( DeltaTime < 0 )
    DeltaTime += MILLISECONDS_PER_DAY;
  OldTime = NewTime;
  
  for (i = 0; i < Timers.Length; ++i) {
    if ( Timers[i].Owner != None ) {
      Timers[i].CurrentTime += DeltaTime;
      if ( Timers[i].CurrentTime > Timers[i].TimerInterval ) {
        Timers[i].Owner.Timer();
        if ( Timers[i].bRepeat ) {
          if ( Timers[i].bFixPrecisionError )
            Timers[i].CurrentTime = Timers[i].CurrentTime % Timers[i].TimerInterval;
          else
            Timers[i].CurrentTime = 0;
        }
        else
          Timers.Remove(i--, 1);
      }
    }
    else
      Timers.Remove(i--, 1);
  }
}


//== static StartTimer ========================================================
/**
Starts a timer.
*/
//=============================================================================

static final function StartTimer(Actor Sender, float TimerInterval, bool bRepeat, optional bool bFixPrecisionError)
{
  local int i;
  local UTP_RealTime theTimer;
  
  if ( Sender == None )
    return;
  theTimer = FindRealTime(Sender);
  if ( theTimer == None )
    return;
  
  if ( TimerInterval > 0 && Sender != None ) {
    TimerInterval *= 1000;
    for (i = 0; i < theTimer.Timers.Length; ++i)
      if ( theTimer.Timers[i].Owner == Sender ) {
        theTimer.Timers[i].TimerInterval = TimerInterval;
        theTimer.Timers[i].CurrentTime = 0;
        theTimer.Timers[i].bRepeat = bRepeat;
        theTimer.Timers[i].bFixPrecisionError = bFixPrecisionError;
        return;
      }
    for (i = 0; i < theTimer.Timers.Length; ++i)
      if ( theTimer.Timers[i].Owner == None )
        break;
    if ( i == theTimer.Timers.Length )
      theTimer.Timers.Insert(i, 1);
    theTimer.Timers[i].TimerInterval = TimerInterval;
    theTimer.Timers[i].CurrentTime = 0;
    theTimer.Timers[i].bRepeat = bRepeat;
    theTimer.Timers[i].bFixPrecisionError = bFixPrecisionError;
    theTimer.Timers[i].Owner = Sender;
  }
  else if ( Sender != None )
    StopTimer(Sender);
}


//== static StopTimer =========================================================
/**
Stops a timer.
*/
//=============================================================================

static final function StopTimer(Actor Sender)
{
  local int i;
  local UTP_RealTime theTimer;
  
  if ( Sender == None )
    return;
  theTimer = FindRealTime(Sender);
  if ( theTimer == None )
    return;
  
  for (i = theTimer.Timers.Length - 1; i >= 0; --i)
    if ( theTimer.Timers[i].Owner == Sender || theTimer.Timers[i].Owner == None )
      theTimer.Timers.Remove(i, 1);
}


//== static SetStopwatch ======================================================
/**
Initializes a stopwatch time index and returns the stopwatch index.
*/
//=============================================================================

static final function int SetStopwatch(Actor Sender)
{
  local int i;
  local UTP_RealTime theTimer;
  
  if ( Sender == None )
    return -1;
  theTimer = FindRealTime(Sender);
  if ( theTimer == None )
    return -1;
  
  i = theTimer.StopwatchStarts.Length;
  theTimer.StopwatchStarts[i] = Sender.Level.MilliSecond + 1000 * (Sender.Level.Second + 60 * (Sender.Level.Minute
      + 60 * Sender.Level.Hour));
  return i;
}


//== static GetStopwatchTime ==================================================
/**
Returns the number of milliseconds since the stopwatch was started.
*/
//=============================================================================

static final function int GetStopwatchTime(Actor Sender, int i)
{
  local UTP_RealTime theTimer;
  
  if ( Sender == None )
    return 0;
  theTimer = FindRealTime(Sender);
  if ( theTimer == None || i < 0 || i >= theTimer.StopwatchStarts.Length )
    return 0;
  
  i = Sender.Level.MilliSecond + 1000 * (Sender.Level.Second + 60 * (Sender.Level.Minute + 60 * Sender.Level.Hour))
      - theTimer.StopwatchStarts[i];
  if ( i < 0 )
    return i + MILLISECONDS_PER_DAY;
  return i;
}


//== operator % (int, int) ====================================================
/**
Modulo operator for integers.
*/
//=============================================================================

static final operator(18) int % (int x, int y) {return x-(x/y*y);}


//== static FindRealTime ======================================================
/**
Returns an existing UTP_RealTime actor or spawns a new one and returns that.
*/
//=============================================================================

private static final function UTP_RealTime FindRealTime(Actor Referencer)
{
  local UTP_RealTime theTimer;
  
  if ( UTP_RealTime(Referencer) != None )
    return UTP_RealTime(Referencer);
  
  foreach Referencer.DynamicActors(class'UTP_RealTime', theTimer)
    return theTimer;
  
  return Referencer.Spawn(class'UTP_RealTime', None);
}


//=============================================================================
// Default properties
//=============================================================================

defaultproperties
{
  bAlwaysTick = True
  bGameRelevant = True
}