/******************************************************************************
UTPlus
Copyright (c) 2005-2006 by Wormbo

This program is free software; you can redistribute and/or modify 
it under the terms of the Open Unreal Mod License version 1.1.
http://wiki.beyondunreal.com/wiki/OpenUnrealModLicense


Container class for utility functions used by the client.
******************************************************************************/


class UTP_Util extends Object
    abstract;


//=============================================================================
// Structs
//=============================================================================

struct TClientPlayerInfo {
  var int           ID;
  var int           IP;
  var array<string> Names;
};


//=============================================================================
// Variables
//=============================================================================

var(UTPlus) noexport const editconst string Build;
var(UTPlus) noexport const editconst string Copyright;
var(UTPlus) noexport const editconst string FriendlyName;

var private editconst int CrcTable[256];

var const editconst string AllowedFileNameChars;


//== static GetVersionString ==================================================
/**
Returns the UTPlus version and build.
*/
//=============================================================================

static function string GetVersionString(optional bool bNoBuild)
{
  if ( bNoBuild )
    return default.FriendlyName;
  else
    return default.FriendlyName @ "build" @ default.Build;
}


//== static GetCopyrightString ================================================
/**
Returns the UTPlus copyright notice.
*/
//=============================================================================

static function string GetCopyrightString()
{
  return default.Copyright;
}


//== Date =====================================================================
/**
Returns a date formatted as a string.
*/
//=============================================================================

static final function string Date(int Year, int Month, int Day)
{
  return Right("0000" $ Year, 4) $ "-" $ Right("00" $ Month, 2) $ "-" $ Right("00" $ Day, 2);
}


//== Time =====================================================================
/**
Returns a time formatted as a string.
*/
//=============================================================================

static final function string Time(int Hour, int Minute, int Second)
{
  return Right("00" $ Hour, 2) $ ":" $ Right("00" $ Minute, 2) $ ":" $ Right("00" $ Second, 2);
}


//== static ParseFileName =====================================================
/**
Finds placeholders in the specified filename template and replaces them accordingly.
*/
//=============================================================================

static final function string ParseFileName(string Template, Actor Referencer)
{
  Template = Repl(Template, "%d", Date(Referencer.Level.Year, Referencer.Level.Month, Referencer.Level.Day));
  Template = Repl(Template, "%t", Repl(Time(Referencer.Level.Hour, Referencer.Level.Minute, Referencer.Level.Second), ":", "-"));
  Template = Repl(Template, "%g", string(Referencer.Level.GetGameClass().Name));
  Template = Repl(Template, "%m", string(Referencer.Outer));
  return Template;
}


//== static StripCodes ========================================================
/**
Strips color codes from a string.
*/
//=============================================================================

static final function string StripCodes(coerce string InputString)
{
  local int i;
  
  i = InStr(InputString, Chr(27));
  while(i != -1) {  
    InputString = Left(InputString, i) $ Mid(InputString, i + 4);
    i = InStr(InputString, Chr(27));
  }
  return InputString;
}


//== static ColorCode =========================================================
/**
Creates a console color code (ESC, R, G, B) from a color.
*/
//=============================================================================

static final function string ColorCode(color Color)
{
  return Chr(27) $ Chr(Max(Color.R, 1)) $ Chr(Max(Color.G, 1)) $ Chr(Max(Color.B, 1));
}


//== static ToColor ===========================================================
/**
Converts a TSkinColor to Color.
*/
//=============================================================================

static final function color ToColor(UTP_Pawn.TSkinColor SkinColor)
{
  local color C;
  
  C.R = SkinColor.R / 4;
  C.G = SkinColor.G / 4;
  C.B = SkinColor.B / 4;
  C.A = 255;
  
  return C;
}


//== static BinarySearchI =====================================================
/**
Search an int array for an element, return its index or -1 if not found.
Modified from wUtils.wArray (http://wiki.beyondunreal.com/WUtils)
*/
//=============================================================================

static final function int BinarySearchI(out array<int> MyArray, int Search, optional out int Closest)
{
  local int Low, High;
  
  High = MyArray.Length - 1;
  
  while (Low <= High) {
    Closest = (Low + High) / 2;
    if ( MyArray[Closest] == Search )
      return Closest;
    if ( MyArray[Closest] > Search )
      High = Closest - 1;
    else if ( MyArray[Closest] < Search )
      Low = Closest + 1;
  }
  
  if ( MyArray.Length != 0 && Low < MyArray.Length && MyArray[Low] < Search )
    Closest = Low + 1;
  else
    Closest = Low;
  
  return -1;
}


//== static BinarySearchS =====================================================
/**
Search a string array for an element, return its index or -1 if not found.
Modified from wUtils.wArray (http://wiki.beyondunreal.com/WUtils)
*/
//=============================================================================

static final function int BinarySearchS(out array<string> MyArray, string Search, optional out int Closest, optional bool bCaseInsensitive)
{
  local int Low, High;
  
  High = MyArray.Length - 1;
  if ( bCaseInsensitive )
    Search = Caps(Search);
  
  while (Low <= High) {
    Closest = (Low + High) / 2;
    if ( bCaseInsensitive ) {
      if ( MyArray[Closest] ~= Search )
        return Closest;
      if ( Caps(MyArray[Closest]) > Search )
        High = Closest - 1;
      else if ( Caps(MyArray[Closest]) < Search )
        Low = Closest + 1;
    }
    else {
      if ( MyArray[Closest] == Search )
        return Closest;
      if ( MyArray[Closest] > Search )
        High = Closest - 1;
      else if ( MyArray[Closest] < Search )
        Low = Closest + 1;
    }
  }
  
  if ( MyArray.Length != 0 && Low < MyArray.Length && (!bCaseInsensitive && MyArray[Low] < Search || !bCaseInsensitive && Caps(MyArray[Low]) < Search) )
    Closest = Low + 1;
  else
    Closest = Low;
  
  return -1;
}


//== static SortI =============================================================
/**
Sort an int array.
*/
//=============================================================================

static final function SortI(out array<int> MyArray)
{
  SortIArrayPart(MyArray, 0, MyArray.Length - 1);
}


//== static SortIArrayPart ====================================================
/**
Sorts an integer array using optimized Insertion Sort.
This implementation outperforms QuickSort due to its use of dynamic arrays and
BinarySearch for unsorted elements.
*/
//=============================================================================

static final function SortIArrayPart(out array<int> MyArray, int LowerBound, int UpperBound)
{
  local int InsertIndex, RemovedIndex;
  local int High, Closest;
  
  if ( LowerBound < UpperBound ) {
    for (RemovedIndex = LowerBound + 1; RemovedIndex <= UpperBound; ++RemovedIndex) {
      if ( MyArray[RemovedIndex-1] > MyArray[RemovedIndex] ) {
        InsertIndex = 0;
        High = RemovedIndex - 1;
        while (InsertIndex <= High) {
          Closest = (InsertIndex + High) / 2;
          if ( MyArray[Closest] == MyArray[RemovedIndex] ) {
            InsertIndex = Closest;
            break;
          }
          if ( MyArray[Closest] > MyArray[RemovedIndex] )
            High = Closest - 1;
          else if ( MyArray[Closest] < MyArray[RemovedIndex] )
            InsertIndex = Closest + 1;
        }
        if ( InsertIndex < RemovedIndex && MyArray[InsertIndex] < MyArray[RemovedIndex] )
          ++InsertIndex;
      }
      else
        InsertIndex = RemovedIndex;
      
      if ( RemovedIndex != InsertIndex ) {
        MyArray.Insert(InsertIndex, 1);
        MyArray[InsertIndex] = MyArray[RemovedIndex + 1];
        MyArray.Remove(RemovedIndex + 1, 1);
      }
    }
  }
}


//== static SortS =============================================================
/**
Sort a string array.
*/
//=============================================================================

static final function SortS(out array<string> MyArray, optional bool bCaseInsensitive)
{
  SortSArrayPart(MyArray, 0, MyArray.Length - 1, bCaseInsensitive);
}


//== static SortSArrayPart ====================================================
/**
Sorts a string array using optimized Insertion Sort.
This implementation outperforms QuickSort due to its use of dynamic arrays and
BinarySearch for unsorted elements.
*/
//=============================================================================

static final function SortSArrayPart(out array<string> MyArray, int LowerBound, int UpperBound, optional bool bCaseInsensitive)
{
  local int InsertIndex, RemovedIndex;
  local int High, Closest;
  
  if ( LowerBound < UpperBound ) {
    for (RemovedIndex = LowerBound + 1; RemovedIndex <= UpperBound; ++RemovedIndex) {
      if ( bCaseInsensitive && Caps(MyArray[RemovedIndex-1]) > Caps(MyArray[RemovedIndex]) ) {
        InsertIndex = 0;
        High = RemovedIndex - 1;
        while (InsertIndex <= High) {
          Closest = (InsertIndex + High) / 2;
          if ( MyArray[Closest] ~= MyArray[RemovedIndex] ) {
            InsertIndex = Closest;
            break;
          }
          if ( Caps(MyArray[Closest]) > Caps(MyArray[RemovedIndex]) )
            High = Closest - 1;
          else if ( Caps(MyArray[Closest]) < Caps(MyArray[RemovedIndex]) )
            InsertIndex = Closest + 1;
        }
        if ( InsertIndex < RemovedIndex && Caps(MyArray[InsertIndex]) < Caps(MyArray[RemovedIndex]) )
          ++InsertIndex;
      }
      else if ( !bCaseInsensitive && MyArray[RemovedIndex-1] > MyArray[RemovedIndex] ) {
        InsertIndex = 0;
        High = RemovedIndex - 1;
        while (InsertIndex <= High) {
          Closest = (InsertIndex + High) / 2;
          if ( MyArray[Closest] == MyArray[RemovedIndex] ) {
            InsertIndex = Closest;
            break;
          }
          if ( MyArray[Closest] > MyArray[RemovedIndex] )
            High = Closest - 1;
          else if ( MyArray[Closest] < MyArray[RemovedIndex] )
            InsertIndex = Closest + 1;
        }
        if ( InsertIndex < RemovedIndex && MyArray[InsertIndex] < MyArray[RemovedIndex] )
          ++InsertIndex;
      }
      else
        InsertIndex = RemovedIndex;
      
      if ( RemovedIndex != InsertIndex ) {
        MyArray.Insert(InsertIndex, 1);
        MyArray[InsertIndex] = MyArray[RemovedIndex + 1];
        MyArray.Remove(RemovedIndex + 1, 1);
      }
    }
  }
}


//== static CrcInit ===========================================================
// CrcInit
//
// Initializes CrcTable and prepares it for use with Crc.
//=============================================================================

static final function CrcInit()
{
  const CrcPolynomial = 0xedb88320;
  
  local int CrcValue;
  local int IndexBit;
  local int IndexEntry;
  
  if ( Default.CrcTable[0] != 0 )
    return; // already initialized
  
  for (IndexEntry = 0; IndexEntry < 256; ++IndexEntry) {
    CrcValue = IndexEntry;
    
    for (IndexBit = 8; IndexBit > 0; --IndexBit)
      if ((CrcValue & 1) != 0)
        CrcValue = (CrcValue >>> 1) ^ CrcPolynomial;
      else
        CrcValue = CrcValue >>> 1;
    
    Default.CrcTable[IndexEntry] = CrcValue;
  }
}


//== static Crc ===============================================================
/**
Calculates and returns a checksum of the given string. Call CrcInit before.
*/
//=============================================================================

static final function int Crc(coerce string Text)
{
  local int CrcValue;
  local int IndexChar;
  
  CrcValue = 0xffffffff;
  
  for (IndexChar = 0; IndexChar < Len(Text); ++IndexChar)
    CrcValue = (CrcValue >>> 8) ^ Default.CrcTable[Asc(Mid(Text, IndexChar, 1)) ^ (CrcValue & 0xff)];
  
  return CrcValue;
}


//== static CrcB ==============================================================
/**
Calculates and returns a checksum of the given byte array. Call CrcInit before.
*/
//=============================================================================

static final function int CrcB(array<byte> Text)
{
  local int CrcValue;
  local int IndexChar;
  
  CrcValue = 0xffffffff;
  
  for (IndexChar = 0; IndexChar < Text.Length; ++IndexChar)
    CrcValue = (CrcValue >>> 8) ^ Default.CrcTable[Text[IndexChar] ^ (CrcValue & 0xff)];
  
  return CrcValue;
}


//== static RepeatChar ========================================================
/**
Returns a char n times.
*/
//=============================================================================

static final function string RepeatChar(string Char, int n)
{
  local string s;
  
  while (n-- > 0)
    s $= Char;
  
  return s;
}


//== static IDToGUID ==========================================================
/**
Converts a 32-digit hexadecimal string to a GUID struct.
*/
//=============================================================================

static final function GUID IDToGUID(string ID)
{
  local GUID Result;
  
  ID = Repl(ID, "-", "");
  Result.A = HexToInt(Mid(ID,  0, 8));
  Result.B = HexToInt(Mid(ID,  8, 8));
  Result.C = HexToInt(Mid(ID, 16, 8));
  Result.D = HexToInt(Mid(ID, 24, 8));
  
  return Result;
}


//== static HexToInt ==========================================================
/**
Converts a hexadecimal string to integer.
*/
//=============================================================================

static final function int HexToInt(string HexNumber)
{
  local int Result;
  local byte c;
  
  HexNumber = Caps(HexNumber);
  
  while (HexNumber != "") {
    c = Asc(HexNumber);
    HexNumber = Mid(HexNumber, 1);
    Result = Result << 4;
    if ( 57 >= c && c >= 48 )
      Result += c - 48;
    else if ( 70 >= c && c >= 65 )
      Result += c - 55;
    else
      return Result;
  }
  
  return Result;
}


//== static IPToInt ===========================================================
/**
Reads an IP string and returns it as an int.
*/
//=============================================================================

static final function int IPToInt(string IP)
{
  local int i;
  local int IPInt;
  
  i = InStr(IP, ":");
  if ( i != -1 )
    IP = Left(IP, i);
  
  while (IP != "") {
    i = InStr(IP, ".");
    if ( i == -1 )
      i = Len(IP);
    IPInt = (IPInt << 8) | int(Left(IP, i));
    IP = Mid(IP, i + 1);
  }
  return IPInt;
}


//== static IntToIP ===========================================================
/**
Reads an int and returns an IP string.
*/
//=============================================================================

static final function string IntToIP(int IP, optional bool bMaskIP)
{
  local string IPString;
  
  if ( bMaskIP )
    IPString = "*";
  else
    IPString = string(IP & 0xFF);
  IP = IP >>> 8;
  IPString = string(IP & 0xFF) $ "." $ IPString;
  IP = IP >>> 8;
  IPString = string(IP & 0xFF) $ "." $ IPString;
  IP = IP >>> 8;
  IPString = string(IP & 0xFF) $ "." $ IPString;
  
  return IPString;
}


//== static CompressBools =====================================================
/**
Combines up to eight bool values into a byte value.
*/
//=============================================================================

static final function CompressBools(out byte Output, byte NumValues, bool B1, optional bool B2, optional bool B3,
    optional bool B4, optional bool B5, optional bool B6, optional bool B7, optional bool B8)
{
  local byte Result;
  
  if ( B1 ) Result +=   1;
  if ( B2 ) Result +=   2;
  if ( B3 ) Result +=   4;
  if ( B4 ) Result +=   8;
  if ( B5 ) Result +=  16;
  if ( B6 ) Result +=  32;
  if ( B7 ) Result +=  64;
  if ( B8 ) Result += 128;
  
  NumValues = Clamp(NumValues, 0, 8);
  
  Output = ((Output >> NumValues) << NumValues) | Result;
}


//== static GetCompressedBool =================================================
/**
Extracts one bit from a byte variable.
*/
//=============================================================================

static final function bool GetCompressedBool(byte Input, byte Bit)
{
  return bool(Input & (2 ** Bit));
}


//=============================================================================
// Default properties
//=============================================================================

defaultproperties
{
  Build="%%%%-%%-%% %%:%%"
  Copyright="Copyright (c) 2005-2006 by Wormbo"
  FriendlyName="UTPlus Client v1"
  
  AllowedFileNameChars = "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_!%()#-"
}
