/******************************************************************************
UTPlus
Copyright (c) 2005-2006 by Wormbo

This program is free software; you can redistribute and/or modify 
it under the terms of the Open Unreal Mod License version 1.1.
http://wiki.beyondunreal.com/wiki/OpenUnrealModLicense


Smoothly cross-fades between multiple values.
******************************************************************************/


class UTP_ValueFader extends Info;


//=============================================================================
// Structs
//=============================================================================

struct TFloatFadeParams {
  var array<float> Values;
  var array<float> Weight;
  var bool bUseSmerp;
  var float Duration;
  var bool bAlwaysTick;
};

struct TVectorFadeParams {
  var array<vector> Values;
  var array<float> Weight;
  var bool bUseSmerp;
  var float Duration;
  var bool bAlwaysTick;
};

struct TRotatorFadeParams {
  var array<rotator> Values;
  var array<float> Weight;
  var bool bUseSmerp;
  var bool bShortestRotation;
  var float Duration;
  var bool bAlwaysTick;
};

struct TColorFadeParams {
  var array<color> Values;
  var array<float> Weight;
  var bool bUseSmerp;
  var float Duration;
  var bool bAlwaysTick;
};


//=============================================================================
// Variables
//=============================================================================

var array<TFloatFadeParams>   FloatFadeParams;
var array<TVectorFadeParams>  VectorFadeParams;
var array<TRotatorFadeParams> RotatorFadeParams;
var array<TColorFadeParams>   ColorFadeParams;


//== static GetValueFader =====================================================
/**
Returns a UTP_ValueFader. A new one will be spawned if there isn't any available yet.
*/
//=============================================================================

static function UTP_ValueFader GetValueFader(Actor Referencer)
{
  local UTP_ValueFader Fader;
  
  if ( Referencer == None )
    return None;
  
  foreach Referencer.DynamicActors(class'UTP_ValueFader', Fader) {
    return Fader;
  }
  return Referencer.Spawn(class'UTP_ValueFader');
}


//== AddFloatFader ============================================================
/**
Adds a float fader. Returns the new fader's ID.
*/
//=============================================================================

function int AddFloatFader(float InitialValue, optional float FadeDuration, optional bool bAlwaysTick, optional bool bUseSmerp)
{
  local int i;
  
  if ( FadeDuration == 0 )
    FadeDuration = 1;
  
  while (i < FloatFadeParams.Length) {
    if ( FloatFadeParams[i].Weight.Length == 0 )
      break;
    ++i;
  }
  if ( i == FloatFadeParams.Length )
    FloatFadeParams.Length = i + 1;
  
  FloatFadeParams[i].Values.Length = 1;
  FloatFadeParams[i].Values[0] = InitialValue;
  FloatFadeParams[i].Weight.Length = 1;
  FloatFadeParams[i].Weight[0] = 1;
  FloatFadeParams[i].bUseSmerp = bUseSmerp;
  FloatFadeParams[i].Duration = FadeDuration;
  FloatFadeParams[i].bAlwaysTick = bAlwaysTick;
  
  return i;
}


//== AddVectorFader ===========================================================
/**
Adds a vector fader. Returns the new fader's ID.
*/
//=============================================================================

function int AddVectorFader(vector InitialValue, optional float FadeDuration, optional bool bAlwaysTick, optional bool bUseSmerp)
{
  local int i;
  
  if ( FadeDuration == 0 )
    FadeDuration = 1;
  
  while (i < VectorFadeParams.Length) {
    if ( VectorFadeParams[i].Weight.Length == 0 )
      break;
    ++i;
  }
  if ( i == VectorFadeParams.Length )
    VectorFadeParams.Length = i + 1;
  
  VectorFadeParams[i].Values.Length = 1;
  VectorFadeParams[i].Values[0] = InitialValue;
  VectorFadeParams[i].Weight.Length = 1;
  VectorFadeParams[i].Weight[0] = 1;
  VectorFadeParams[i].bUseSmerp = bUseSmerp;
  VectorFadeParams[i].Duration = FadeDuration;
  VectorFadeParams[i].bAlwaysTick = bAlwaysTick;
  
  return i;
}


//== AddRotatorFader ==========================================================
/**
Adds a rotator fader. Returns the new fader's ID.
*/
//=============================================================================

function int AddRotatorFader(rotator InitialValue, optional float FadeDuration, optional bool bAlwaysTick, optional bool bUseSmerp, optional bool bShortestRotation)
{
  local int i;
  
  if ( FadeDuration == 0 )
    FadeDuration = 1;
  
  while (i < RotatorFadeParams.Length) {
    if ( RotatorFadeParams[i].Weight.Length == 0 )
      break;
    ++i;
  }
  if ( i == RotatorFadeParams.Length )
    RotatorFadeParams.Length = i + 1;
  
  if ( bShortestRotation )
    InitialValue = Normalize(InitialValue);
  
  RotatorFadeParams[i].Values.Length = 1;
  RotatorFadeParams[i].Values[0] = InitialValue;
  RotatorFadeParams[i].Weight.Length = 1;
  RotatorFadeParams[i].Weight[0] = 1;
  RotatorFadeParams[i].bUseSmerp = bUseSmerp;
  RotatorFadeParams[i].bShortestRotation = bShortestRotation;
  RotatorFadeParams[i].Duration = FadeDuration;
  RotatorFadeParams[i].bAlwaysTick = bAlwaysTick;
  
  return i;
}


//== AddColorFader ==========================================================
/**
Adds a color fader. Returns the new fader's ID.
*/
//=============================================================================

function int AddColorFader(color InitialValue, optional float FadeDuration, optional bool bAlwaysTick, optional bool bUseSmerp)
{
  local int i;
  
  if ( FadeDuration == 0 )
    FadeDuration = 1;
  
  while (i < ColorFadeParams.Length) {
    if ( ColorFadeParams[i].Weight.Length == 0 )
      break;
    ++i;
  }
  if ( i == ColorFadeParams.Length )
    ColorFadeParams.Length = i + 1;
  
  ColorFadeParams[i].Values.Length = 1;
  ColorFadeParams[i].Values[0] = InitialValue;
  ColorFadeParams[i].Weight.Length = 1;
  ColorFadeParams[i].Weight[0] = 1;
  ColorFadeParams[i].bUseSmerp = bUseSmerp;
  ColorFadeParams[i].Duration = FadeDuration;
  ColorFadeParams[i].bAlwaysTick = bAlwaysTick;
  
  return i;
}


//== AddFloatFadeValue ========================================================
/**
Adds a new value to fade to.
*/
//=============================================================================

function AddFloatFadeValue(int FaderNum, float NewValue, optional bool bUpdateLastValue)
{
  if ( FaderNum < 0 || FaderNum >= FloatFadeParams.Length || FloatFadeParams[FaderNum].Values.Length == 0 ) {
    warn("Float FaderNum out of bounds:"@FaderNum$"/"$FloatFadeParams.Length);
    return;
  }
  if ( bUpdateLastValue ) {
    FloatFadeParams[FaderNum].Values[FloatFadeParams[FaderNum].Values.Length - 1] = NewValue;
  }
  else {
    FloatFadeParams[FaderNum].Values[FloatFadeParams[FaderNum].Values.Length] = NewValue;
    FloatFadeParams[FaderNum].Weight.Length = FloatFadeParams[FaderNum].Values.Length;
  }
}


//== AddVectorFadeValue =======================================================
/**
Adds a new value to fade to.
*/
//=============================================================================

function AddVectorFadeValue(int FaderNum, vector NewValue, optional bool bUpdateLastValue)
{
  if ( FaderNum < 0 || FaderNum >= VectorFadeParams.Length || VectorFadeParams[FaderNum].Values.Length == 0 ) {
    warn("Vector FaderNum out of bounds:"@FaderNum$"/"$VectorFadeParams.Length);
    return;
  }
  
  if ( bUpdateLastValue ) {
    VectorFadeParams[FaderNum].Values[VectorFadeParams[FaderNum].Values.Length - 1] = NewValue;
  }
  else {
    VectorFadeParams[FaderNum].Values[VectorFadeParams[FaderNum].Values.Length] = NewValue;
    VectorFadeParams[FaderNum].Weight.Length = VectorFadeParams[FaderNum].Values.Length;
  }
}


//== AddRotatorFadeValue ======================================================
/**
Adds a new value to fade to.
*/
//=============================================================================

function AddRotatorFadeValue(int FaderNum, rotator NewValue, optional bool bUpdateLastValue)
{
  if ( FaderNum < 0 || FaderNum >= RotatorFadeParams.Length || RotatorFadeParams[FaderNum].Values.Length == 0 ) {
    warn("Rotator FaderNum out of bounds:"@FaderNum$"/"$RotatorFadeParams.Length);
    return;
  }
  if ( RotatorFadeParams[FaderNum].bShortestRotation )
    NewValue = Normalize(NewValue);
  if ( bUpdateLastValue ) {
    RotatorFadeParams[FaderNum].Values[RotatorFadeParams[FaderNum].Values.Length - 1] = NewValue;
  }
  else {
    RotatorFadeParams[FaderNum].Values[RotatorFadeParams[FaderNum].Values.Length] = NewValue;
    RotatorFadeParams[FaderNum].Weight.Length = RotatorFadeParams[FaderNum].Values.Length;
  }
}


//== AddColorFadeValue ========================================================
/**
Adds a new value to fade to.
*/
//=============================================================================

function AddColorFadeValue(int FaderNum, color NewValue, optional bool bUpdateLastValue)
{
  if ( FaderNum < 0 || FaderNum >= ColorFadeParams.Length || ColorFadeParams[FaderNum].Values.Length == 0 ) {
    warn("Color FaderNum out of bounds:"@FaderNum$"/"$ColorFadeParams.Length);
    return;
  }
  if ( bUpdateLastValue ) {
    ColorFadeParams[FaderNum].Values[ColorFadeParams[FaderNum].Values.Length - 1] = NewValue;
  }
  else {
    ColorFadeParams[FaderNum].Values[ColorFadeParams[FaderNum].Values.Length] = NewValue;
    ColorFadeParams[FaderNum].Weight.Length = ColorFadeParams[FaderNum].Values.Length;
  }
}


//== RemoveFloatFader ========================================================
/**
Removes the specified float fader.
*/
//=============================================================================

function RemoveFloatFader(int FaderNum)
{
  local int i;
  
  if ( FaderNum < 0 || FaderNum >= FloatFadeParams.Length || FloatFadeParams[FaderNum].Values.Length == 0 ) {
    warn("Float FaderNum doesn't exist:" @ FaderNum $"/"$ FloatFadeParams.Length);
    return;
  }
  FloatFadeParams[FaderNum].Values.Length = 0;
  FloatFadeParams[FaderNum].Weight.Length = 0;
  
  for (i = FloatFadeParams.Length - 1; i >= 0; --i) {
    if ( FloatFadeParams[i].Values.Length > 0 )
      break;
    FloatFadeParams.Remove(i, 1);
  }
}


//== RemoveVectorFader ========================================================
/**
Removes the specified vector fader.
*/
//=============================================================================

function RemoveVectorFader(int FaderNum)
{
  local int i;
  
  if ( FaderNum < 0 || FaderNum >= VectorFadeParams.Length || VectorFadeParams[FaderNum].Values.Length == 0 ) {
    warn("Vector FaderNum doesn't exist:" @ FaderNum $"/"$ VectorFadeParams.Length);
    return;
  }
  
  VectorFadeParams[FaderNum].Values.Length = 0;
  VectorFadeParams[FaderNum].Weight.Length = 0;
  
  for (i = VectorFadeParams.Length - 1; i >= 0; --i) {
    if ( VectorFadeParams[i].Values.Length > 0 )
      break;
    VectorFadeParams.Remove(i, 1);
  }
}


//== RemoveRotatorFader ========================================================
/**
Removes the specified rotator fader.
*/
//=============================================================================

function RemoveRotatorFader(int FaderNum)
{
  local int i;
  
  if ( FaderNum < 0 || FaderNum >= RotatorFadeParams.Length || RotatorFadeParams[FaderNum].Values.Length == 0 ) {
    warn("Rotator FaderNum doesn't exist:" @ FaderNum $"/"$ RotatorFadeParams.Length);
    return;
  }
  RotatorFadeParams[FaderNum].Values.Length = 0;
  RotatorFadeParams[FaderNum].Weight.Length = 0;
  
  for (i = RotatorFadeParams.Length - 1; i >= 0; --i) {
    if ( RotatorFadeParams[i].Values.Length > 0 )
      break;
    RotatorFadeParams.Remove(i, 1);
  }
}


//== RemoveColorFader ========================================================
/**
Removes the specified color fader.
*/
//=============================================================================

function RemoveColorFader(int FaderNum)
{
  local int i;
  
  if ( FaderNum < 0 || FaderNum >= ColorFadeParams.Length || ColorFadeParams[FaderNum].Values.Length == 0 ) {
    warn("Color FaderNum doesn't exist:" @ FaderNum $"/"$ ColorFadeParams.Length);
    return;
  }
  ColorFadeParams[FaderNum].Values.Length = 0;
  ColorFadeParams[FaderNum].Weight.Length = 0;
  
  for (i = ColorFadeParams.Length - 1; i >= 0; --i) {
    if ( ColorFadeParams[i].Values.Length > 0 )
      break;
    ColorFadeParams.Remove(i, 1);
  }
}


//== GetFadedFloatValue =======================================================
/**
Returns the specified float fader's value.
*/
//=============================================================================

function float GetFadedFloatValue(int FaderNum)
{
  local float FadedValue;
  local float Weight, TotalWeight;
  local int i;
  
  if ( FaderNum < 0 || FaderNum >= FloatFadeParams.Length || FloatFadeParams[FaderNum].Values.Length == 0 ) {
    warn("Float FaderNum not available:" @ FaderNum $ "/" $ FloatFadeParams.Length);
    return FadedValue;
  }
  for (i = 0; i < FloatFadeParams[FaderNum].Values.Length; ++i) {
    if ( FloatFadeParams[FaderNum].bUseSmerp )
      Weight = Smerp(FloatFadeParams[FaderNum].Weight[i], 0, 1);
    else
      Weight = FloatFadeParams[FaderNum].Weight[i];
    FadedValue += FloatFadeParams[FaderNum].Values[i] * Weight;
    TotalWeight += Weight;
  }
  
  return FadedValue / TotalWeight;
}


//== GetFadedVectorValue =======================================================
/**
Returns the specified vector fader's value.
*/
//=============================================================================

function vector GetFadedVectorValue(int FaderNum)
{
  local vector FadedValue;
  local float Weight, TotalWeight;
  local int i;
  
  if ( FaderNum < 0 || FaderNum >= VectorFadeParams.Length || VectorFadeParams[FaderNum].Values.Length == 0 ) {
    warn("Vector FaderNum not available:" @ FaderNum $ "/" $ VectorFadeParams.Length);
    return FadedValue;
  }
  for (i = 0; i < VectorFadeParams[FaderNum].Values.Length; ++i) {
    if ( VectorFadeParams[FaderNum].bUseSmerp )
      Weight = Smerp(VectorFadeParams[FaderNum].Weight[i], 0, 1);
    else
      Weight = VectorFadeParams[FaderNum].Weight[i];
    FadedValue += VectorFadeParams[FaderNum].Values[i] * Weight;
    TotalWeight += Weight;
  }
  
  return FadedValue / TotalWeight;
}


//== GetFadedRotatorValue =======================================================
/**
Returns the specified rotator fader's value.
*/
//=============================================================================

function rotator GetFadedRotatorValue(int FaderNum)
{
  local rotator FadedValue, TempValue;
  local float FadedPitch, FadedYaw, FadedRoll;
  local float Weight, TotalWeight;
  local int i, DeltaRot;
  
  if ( FaderNum < 0 || FaderNum >= RotatorFadeParams.Length || RotatorFadeParams[FaderNum].Values.Length == 0 ) {
    warn("Rotator FaderNum not available:" @ FaderNum $ "/" $ RotatorFadeParams.Length);
    return FadedValue;
  }
  if ( RotatorFadeParams[FaderNum].bShortestRotation ) {
    FadedValue = RotatorFadeParams[FaderNum].Values[0];
    for (i = 1; i < RotatorFadeParams[FaderNum].Values.Length; ++i) {
      TempValue = RotatorFadeParams[FaderNum].Values[i];
      
      if ( RotatorFadeParams[FaderNum].bUseSmerp )
        Weight = Smerp(RotatorFadeParams[FaderNum].Weight[i], 0, 1);
      else
        Weight = RotatorFadeParams[FaderNum].Weight[i];
      
      DeltaRot = (TempValue.Yaw & 65535) - (FadedValue.Yaw & 65535);
      if (DeltaRot < -32768)
        DeltaRot += 65536;
      else if (DeltaRot > 32768)
        DeltaRot -= 65536;
			FadedValue.Yaw += Weight * DeltaRot;
      
      DeltaRot = (TempValue.Pitch & 65535) - (FadedValue.Pitch & 65535);
      if (DeltaRot < -32768)
        DeltaRot += 65536;
      else if (DeltaRot > 32768)
        DeltaRot -= 65536;
			FadedValue.Pitch += Weight * DeltaRot;
      
      DeltaRot = (TempValue.Roll & 65535) - (FadedValue.Roll & 65535);
      if ( DeltaRot < -32768 )
        DeltaRot += 65536;
      else if ( DeltaRot > 32768 )
        DeltaRot -= 65536;
			FadedValue.Roll += Weight * DeltaRot;
    }
    FadedValue = Normalize(FadedValue);
  }
  else {
    for (i = 0; i < RotatorFadeParams[FaderNum].Values.Length; ++i) {
      if ( RotatorFadeParams[FaderNum].bUseSmerp )
        Weight = Smerp(RotatorFadeParams[FaderNum].Weight[i], 0, 1);
      else
        Weight = RotatorFadeParams[FaderNum].Weight[i];
      FadedPitch  += RotatorFadeParams[FaderNum].Values[i].Pitch  * Weight;
      FadedYaw    += RotatorFadeParams[FaderNum].Values[i].Yaw    * Weight;
      FadedRoll   += RotatorFadeParams[FaderNum].Values[i].Roll   * Weight;
      TotalWeight += Weight;
    }
    
    FadedValue.Pitch  = FadedPitch / TotalWeight;
    FadedValue.Yaw    = FadedYaw   / TotalWeight;
    FadedValue.Roll   = FadedRoll  / TotalWeight;
  }
  return FadedValue;
}


//== GetFadedColorValue =======================================================
/**
Returns the specified color fader's value.
*/
//=============================================================================

function color GetFadedColorValue(int FaderNum)
{
  local color FadedValue;
  local float FadedR, FadedG, FadedB, FadedA;
  local float Weight, TotalWeight;
  local int i;
  
  if ( FaderNum < 0 || FaderNum >= ColorFadeParams.Length || ColorFadeParams[FaderNum].Values.Length == 0 ) {
    warn("Color FaderNum not available:" @ FaderNum $ "/" $ ColorFadeParams.Length);
    return FadedValue;
  }
  for (i = 0; i < ColorFadeParams[FaderNum].Values.Length; ++i) {
    if ( ColorFadeParams[FaderNum].bUseSmerp )
      Weight = Smerp(ColorFadeParams[FaderNum].Weight[i], 0, 1);
    else
      Weight = ColorFadeParams[FaderNum].Weight[i];
    FadedR += ColorFadeParams[FaderNum].Values[i].R * Weight;
    FadedG += ColorFadeParams[FaderNum].Values[i].G * Weight;
    FadedB += ColorFadeParams[FaderNum].Values[i].B * Weight;
    FadedA += ColorFadeParams[FaderNum].Values[i].A * Weight;
    TotalWeight += Weight;
  }
  
  FadedValue.R = FClamp(FadedR / TotalWeight, 0, 255);
  FadedValue.G = FClamp(FadedG / TotalWeight, 0, 255);
  FadedValue.B = FClamp(FadedB / TotalWeight, 0, 255);
  FadedValue.A = FClamp(FadedA / TotalWeight, 0, 255);
  
  return FadedValue;
}


//== Tick =====================================================================
/**
Update all faders.
*/
//=============================================================================

event Tick(float DeltaTime)
{
  local bool bPaused;
  local int i, j;
  
  bPaused = Level.Pauser != None;
  
  for (i = 0; i < FloatFadeParams.Length; ++i) {
    if ( FloatFadeParams[i].Weight.Length > 0 && (FloatFadeParams[i].bAlwaysTick || !bPaused) ) {
      for (j = 0; j < FloatFadeParams[i].Weight.Length - 1; ++j) {
        FloatFadeParams[i].Weight[j] -= DeltaTime / FloatFadeParams[i].Duration;
        if ( FloatFadeParams[i].Weight[j] <= 0 ) {
          FloatFadeParams[i].Values.Remove(j, 1);
          FloatFadeParams[i].Weight.Remove(j--, 1);
        }
      }
      if ( FloatFadeParams[i].Weight[j] < 1 ) {
        FloatFadeParams[i].Weight[j] += DeltaTime / FloatFadeParams[i].Duration;
        if ( FloatFadeParams[i].Weight[j] > 1 )
          FloatFadeParams[i].Weight[j] = 1;
      }
    }
  }
  
  for (i = 0; i < VectorFadeParams.Length; ++i) {
    if ( VectorFadeParams[i].Weight.Length > 0 && (VectorFadeParams[i].bAlwaysTick || !bPaused) ) {
      for (j = 0; j < VectorFadeParams[i].Weight.Length - 1; ++j) {
        VectorFadeParams[i].Weight[j] -= DeltaTime / VectorFadeParams[i].Duration;
        if ( VectorFadeParams[i].Weight[j] <= 0 ) {
          VectorFadeParams[i].Values.Remove(j, 1);
          VectorFadeParams[i].Weight.Remove(j--, 1);
        }
      }
      if ( VectorFadeParams[i].Weight[j] < 1 ) {
        VectorFadeParams[i].Weight[j] += DeltaTime / VectorFadeParams[i].Duration;
        if ( VectorFadeParams[i].Weight[j] > 1 )
          VectorFadeParams[i].Weight[j] = 1;
      }
    }
  }
  
  for (i = 0; i < RotatorFadeParams.Length; ++i) {
    if ( RotatorFadeParams[i].Weight.Length > 0 && (RotatorFadeParams[i].bAlwaysTick || !bPaused) ) {
      for (j = 0; j < RotatorFadeParams[i].Weight.Length - 1; ++j) {
        RotatorFadeParams[i].Weight[j] -= DeltaTime / RotatorFadeParams[i].Duration;
        if ( RotatorFadeParams[i].Weight[j] <= 0 ) {
          RotatorFadeParams[i].Values.Remove(j, 1);
          RotatorFadeParams[i].Weight.Remove(j--, 1);
        }
      }
      if ( RotatorFadeParams[i].Weight[j] < 1 ) {
        RotatorFadeParams[i].Weight[j] += DeltaTime / RotatorFadeParams[i].Duration;
        if ( RotatorFadeParams[i].Weight[j] > 1 )
          RotatorFadeParams[i].Weight[j] = 1;
      }
    }
  }
  
  for (i = 0; i < ColorFadeParams.Length; ++i) {
    if ( ColorFadeParams[i].Weight.Length > 0 && (ColorFadeParams[i].bAlwaysTick || !bPaused) ) {
      for (j = 0; j < ColorFadeParams[i].Weight.Length - 1; ++j) {
        ColorFadeParams[i].Weight[j] -= DeltaTime / ColorFadeParams[i].Duration;
        if ( ColorFadeParams[i].Weight[j] <= 0 ) {
          ColorFadeParams[i].Values.Remove(j, 1);
          ColorFadeParams[i].Weight.Remove(j--, 1);
        }
      }
      if ( ColorFadeParams[i].Weight[j] < 1 ) {
        ColorFadeParams[i].Weight[j] += DeltaTime / ColorFadeParams[i].Duration;
        if ( ColorFadeParams[i].Weight[j] > 1 )
          ColorFadeParams[i].Weight[j] = 1;
      }
    }
  }
}


//=============================================================================
// Default properties
//=============================================================================

defaultproperties
{
  bAlwaysTick = True
}
