/******************************************************************************
UTPlus
Copyright (c) 2005-2006 by Wormbo

This program is free software; you can redistribute and/or modify 
it under the terms of the Open Unreal Mod License version 1.1.
http://wiki.beyondunreal.com/wiki/OpenUnrealModLicense


Enables advanced movement features.
The class name does not include "UTPlus" so the mutator can be filtered
together with the stand-alone MultiDodging mutator without requiring an
additional server filter.
******************************************************************************/


class MutMultiDodging extends Mutator
  config;


//=============================================================================
// Configuration
//=============================================================================

var config int  MaxDodgesPerJump;
var config bool bAlwaysAllowDodgeDoubleJump;
var config bool bEnableBoostDodge;
var config bool bFastDodging;
var config bool bImprovedSlopeDodging;
var config bool bPawnsIgnoreTerminalVelocity;


//=============================================================================
// Localization
//=============================================================================

var localized string lblMaxDodgesPerJump;
var localized string lblAlwaysAllowDodgeDoubleJump;
var localized string lblEnableBoostDodge;
var localized string lblFastDodging;
var localized string lblImprovedSlopeDodging;
var localized string lblPawnsIgnoreTerminalVelocity;

var localized string descMaxDodgesPerJump;
var localized string descAlwaysAllowDodgeDoubleJump;
var localized string descEnableBoostDodge;
var localized string descFastDodging;
var localized string descImprovedSlopeDodging;
var localized string descPawnsIgnoreTerminalVelocity;


//== PostBeginPlay ============================================================
/**
Enable advanced movement options in the UTP_RI.
*/
//=============================================================================

function PostBeginPlay()
{
  local UTP_ReplicationInfo UTPlusRI;
  
  if ( Level.Game.DefaultPlayerClassName ~= string(class'xPawn') || Level.Game.DefaultPlayerClassName ~= string(class'xMutantPawn') )
    Level.Game.DefaultPlayerClassName = string(class'UTP_Pawn');
  
  UTPlusRI = class'UTP_ReplicationInfo'.static.InitRI(Self);
  UTPlusRI.bMultiDodgingEnabled         = True;
  UTPlusRI.bBoostDodgeEnabled           = bEnableBoostDodge;  // only for information
  UTPlusRI.MaxDodgesPerJump             = MaxDodgesPerJump;
  UTPlusRI.bAlwaysAllowDodgeDoubleJump  = bAlwaysAllowDodgeDoubleJump;
  UTPlusRI.bFastDodging                 = bFastDodging;
  UTPlusRI.bImprovedSlopeDodging        = bImprovedSlopeDodging;
  UTPlusRI.bPawnsIgnoreTerminalVelocity = bPawnsIgnoreTerminalVelocity;
}


//== CheckReplacement =========================================================
/**
Replace the pawn class for all controllers.
*/
//=============================================================================

function bool CheckReplacement(Actor Other, out byte bSuperRelevant)
{
  if ( Controller(Other) != None && Controller(Other).bIsPlayer && (Controller(Other).PawnClass == None || Controller(Other).PawnClass == class'xPawn' || Controller(Other).PawnClass == class'xMutantPawn') ) {
    //log("CheckReplacement"@Other);
    Controller(Other).PawnClass = class'UTP_Pawn';
    Controller(Other).PreviousPawnClass = class'UTP_Pawn';  // force PlayerChangedClass() in Mutant and for bots
  }
  return Super.CheckReplacement(Other, bSuperRelevant);
}


//== PlayerChangedClass =======================================================
/**
Make sure the players keep the UTP_Pawn as PawnClass.
*/
//=============================================================================

function PlayerChangedClass(Controller thePlayer)
{
  Super.PlayerChangedClass(thePlayer);
  
  if ( thePlayer.PreviousPawnClass == class'UTP_Pawn' ) {
    if ( thePlayer.PawnClass == class'xPawn' || thePlayer.PawnClass == class'xMutantPawn' )
      thePlayer.PawnClass = class'UTP_Pawn';
  }
}


//== ModifyPlayer =============================================================
/**
Enables boost-dodging, disables terminal velocity server-side and makes sure there is a UTP_PlayerReplicationInfo for the player.
*/
//=============================================================================

function ModifyPlayer(Pawn Other)
{
  Super.ModifyPlayer(Other);
  
  if ( xPawn(Other) != None ) {
    if ( UTP_Pawn(Other) == None )
      class'UTP_PlayerReplicationInfo'.static.FindFor(Other);
    else if ( bImprovedSlopeDodging )
      UTP_Pawn(Other).bImprovedSlopeDodging = True;
    if ( bEnableBoostDodge )
      xPawn(Other).bCanBoostDodge = True;
    if ( bPawnsIgnoreTerminalVelocity )
      Other.bIgnoreTerminalVelocity = True;
  }
}


//== FillPlayInfo =============================================================
/**
Adds the Multi-Dodging configuration to the webadmin interface.
*/
//=============================================================================

static function FillPlayInfo(PlayInfo PlayInfo)
{
  if ( class'UTP_ServerUtil'.static.ClassIsInPlayInfo(PlayInfo, default.Class) )
    return;
  
  Super.FillPlayInfo(PlayInfo);
  
  PlayInfo.AddSetting(class'MutUTPlus'.default.FriendlyName, "MaxDodgesPerJump",              default.lblMaxDodgesPerJump,            0, 60, "Text", "3;-1:999");
  PlayInfo.AddSetting(class'MutUTPlus'.default.FriendlyName, "bAlwaysAllowDodgeDoubleJump",   default.lblAlwaysAllowDodgeDoubleJump,  0, 61, "Check");
  PlayInfo.AddSetting(class'MutUTPlus'.default.FriendlyName, "bEnableBoostDodge",             default.lblEnableBoostDodge,            0, 62, "Check");
  PlayInfo.AddSetting(class'MutUTPlus'.default.FriendlyName, "bFastDodging",                  default.lblFastDodging,                 0, 63, "Check");
  PlayInfo.AddSetting(class'MutUTPlus'.default.FriendlyName, "bImprovedSlopeDodging",         default.lblImprovedSlopeDodging,        0, 64, "Check");
  PlayInfo.AddSetting(class'MutUTPlus'.default.FriendlyName, "bPawnsIgnoreTerminalVelocity",  default.lblPawnsIgnoreTerminalVelocity, 0, 65, "Check");
  
  PlayInfo.PopClass();
}


//== GetDescriptionText =======================================================
/**
Returns a description text for the specified property.
*/
//=============================================================================

static event string GetDescriptionText(string PropName)
{
  Switch (PropName) {
    Case "MaxDodgesPerJump":
      return default.descMaxDodgesPerJump;
    Case "bAlwaysAllowDodgeDoubleJump":
      return default.descAlwaysAllowDodgeDoubleJump;
    Case "bEnableBoostDodge":
      return default.descEnableBoostDodge;
    Case "bFastDodging":
      return default.descFastDodging;
    Case "bImprovedSlopeDodging":
      return default.descImprovedSlopeDodging;
    Case "bPawnsIgnoreTerminalVelocity":
      return default.descPawnsIgnoreTerminalVelocity;
  }
  return Super.GetDescriptionText(PropName);
}


//=============================================================================
// Default properties
//=============================================================================

defaultproperties
{
  GroupName    = "Dodging"
  FriendlyName = "UTPlus v1 Advanced Movement"
  Description  = "Enables advanced movement features like multiple wall dodges per jump (also known as 'Multi-Dodging'), faster dodging after landing, boost-dodging and always allowing a double-jump after dodging."
  
  MaxDodgesPerJump              = -1
  bAlwaysAllowDodgeDoubleJump   = True
  bEnableBoostDodge             = False
  bFastDodging                  = True
  bImprovedSlopeDodging         = True
  bPawnsIgnoreTerminalVelocity  = True
  
  lblMaxDodgesPerJump             = "Max. additional dodges per jump"
  lblAlwaysAllowDodgeDoubleJump   = "Always allow dodge-doublejumps"
  lblEnableBoostDodge             = "Enable boost dodge"
  lblFastDodging                  = "Fast dodging"
  lblImprovedSlopeDodging         = "Improved slope dodging"
  lblPawnsIgnoreTerminalVelocity  = "Disable movement speed limitation"
  
  descMaxDodgesPerJump            = "The maximum number of additional dodges per jumping move. (-1 = unlimited)"
  descAlwaysAllowDodgeDoubleJump  = "Whether every dodging move can be followed up by a double-jump."
  descEnableBoostDodge            = "Adds a jump's Z velocity to the wall dodge's upward velocity like in UT2003. Allowes for much greater distance if jump and walldodge are executed simultaneously but also cuts off a series of walldodges much earlier due to the increasing downward velocity."
  descFastDodging                 = "Allowes players to buffer in the first click of a dodge double click while they are still in the air. (Only works with at least one walldodge remaining for the jump.)"
  descImprovedSlopeDodging        = "Adjusts the dodging direction when dodging up a (walkable) slope."
  descPawnsIgnoreTerminalVelocity = "Disables the upper limit of 2500 UU/s for a player's movement speed while jumping."
}