/******************************************************************************
UTPlus
Copyright (c) 2005-2006 by Wormbo

This program is free software; you can redistribute and/or modify 
it under the terms of the Open Unreal Mod License version 1.1.
http://wiki.beyondunreal.com/wiki/OpenUnrealModLicense


Tries to compensate ping for network clients.
The class name does not include "UTPlus" so the mutator can be filtered
together with the stand-alone ping compensation mutator without requiring an
additional server filter.
******************************************************************************/


class MutPingCompensation extends Mutator;


//=============================================================================
// Variables
//=============================================================================

var bool bUnlagged;
var UTP_ReplicationInfo UTPlusRI;

var array<UTP_Pawn> Pawns;


//== PostBeginPlay ============================================================
/**
Find the UTP replication info.
*/
//=============================================================================

function PostBeginPlay()
{
  local Mutator M;
  
  Super.PostBeginPlay();
  
  if ( Level.NetMode != NM_Standalone ) {
    for (M = Level.Game.BaseMutator; M != None; M = M.NextMutator) {
      if (MutInstaGib(M) != None) {
        bUnlagged = True;
        break;
      }
    }
    if ( Level.Game.DefaultPlayerClassName ~= string(class'xPawn') || Level.Game.DefaultPlayerClassName ~= string(class'xMutantPawn') )
      Level.Game.DefaultPlayerClassName = string(class'UTP_Pawn');
    
    UTPlusRI = class'UTP_ReplicationInfo'.static.InitRI(Self);
    UTPlusRI.bAllowPingPrediction = True;
    if (bUnlagged)
      UTPlusRI.bServerPingCompensationAvailable = True;
  }
  else
    log("NOTE:" @ Class @ "not required in single player.", 'UTPlus');
}


//== AddMutator ===============================================================
/**
Look for InstaGib mutator.
*/
//=============================================================================

function AddMutator(Mutator Other)
{
  Super.AddMutator(Other);
  if (UTPlusRI != None && MutInstaGib(Other) != None) {
    bUnlagged = True;
    UTPlusRI.bServerPingCompensationAvailable = True;
  }
}


//== Tick =====================================================================
/**
Make this mutator triggerable.
*/
//=============================================================================

function Tick(float DeltaTime)
{
  if (bUnlagged && Tag != 'UnlagPawns')
    Tag = 'UnlagPawns';
  
  Disable('Tick');
}


//== Trigger ==================================================================
/**
Unlag relevant pawns.
*/
//=============================================================================

event Trigger(Actor Other, Pawn EventInstigator)
{
  local UTP_PlayerReplicationInfo UTPPRI;
  local int i;
  
  if (!UTPlusRI.bServerPingCompensationAvailable)
    return;
  
  UTPPRI = class'UTP_PlayerReplicationInfo'.static.GetFor(EventInstigator);
  if (UTPPRI != None && UTPPRI.GamePing > 0 && (UTPPRI.PingCompensationType > PredictionPreferred || !UTPlusRI.bAllowPingPrediction && UTPPRI.PingCompensationType == PredictionPreferred)) {
    for (i = 0; i < Pawns.Length; i++) {
      if (Pawns[i] == None || Pawns[i].Health <= 0)
        Pawns.Remove(i--, 1);
      else if (Pawns[i] != EventInstigator)
        Pawns[i].EnableUnlag(UTPPRI.GamePing * 0.5);
    }
  }
}


//== Untrigger ================================================================
/**
Restore pawn locations.
*/
//=============================================================================

event Untrigger(Actor Other, Pawn EventInstigator)
{
  local int i;
  
  for (i = 0; i < Pawns.Length; i++) {
    if (Pawns[i] == None || Pawns[i].Health <= 0)
      Pawns.Remove(i--, 1);
    else
      Pawns[i].DisableUnlag();
  }
}


//== CheckReplacement =========================================================
/**
Replace the pawn class for all controllers.
*/
//=============================================================================

function bool AlwaysKeep(Actor Other)
{
  if (bUnlagged) {
    if (Weapon(Other) != None) {
      if (Weapon(Other).FireModeClass[0] == class'SuperShockBeamFire' || Weapon(Other).FireModeClass[0] == class'ZoomSuperShockBeamFire')
        Weapon(Other).FireModeClass[0] = class'UTP_UnlaggedSuperShockBeamFire';
      if (Weapon(Other).FireModeClass[1] == class'SuperShockBeamFire' || Weapon(Other).FireModeClass[1] == class'ZoomSuperShockBeamFire') 
        Weapon(Other).FireModeClass[1] = class'UTP_UnlaggedSuperShockBeamFire';
    }
  }
  return Super.AlwaysKeep(Other);
}

//== CheckReplacement =========================================================
/**
Replace the pawn class for all controllers.
*/
//=============================================================================

function bool CheckReplacement(Actor Other, out byte bSuperRelevant)
{
  if ( Controller(Other) != None && Controller(Other).bIsPlayer && (Controller(Other).PawnClass == None || Controller(Other).PawnClass == class'xPawn' || Controller(Other).PawnClass == class'xMutantPawn') ) {
    Controller(Other).PawnClass = class'UTP_Pawn';
    Controller(Other).PreviousPawnClass = class'UTP_Pawn';  // force PlayerChangedClass() in Mutant and for bots
  }
  if (bUnlagged && UTP_Pawn(Other) != None)
    Pawns[Pawns.Length] = UTP_Pawn(Other);
  
  return Super.CheckReplacement(Other, bSuperRelevant);
}


//== PlayerChangedClass =======================================================
/**
Make sure the players keep the UTP_Pawn as PawnClass.
*/
//=============================================================================

function PlayerChangedClass(Controller thePlayer)
{
  Super.PlayerChangedClass(thePlayer);
  
  if ( thePlayer.PreviousPawnClass == class'UTP_Pawn' ) {
    if ( thePlayer.PawnClass == class'xPawn' || thePlayer.PawnClass == class'xMutantPawn' )
      thePlayer.PawnClass = class'UTP_Pawn';
  }
}


//=============================================================================
// Default properties
//=============================================================================

defaultproperties
{
  FriendlyName  = "UTPlus v1 Ping Compensation"
  Description   = "Allows clients to use ping prediction to compensate the ping effect. When combined with InstaGib, players will be 'unlagged' serversidely."
  GroupName     = "PingCompensation"
}