/******************************************************************************
UTPlus
Copyright (c) 2005-2006 by Wormbo

This program is free software; you can redistribute and/or modify 
it under the terms of the Open Unreal Mod License version 1.1.
http://wiki.beyondunreal.com/wiki/OpenUnrealModLicense


Abstract base class for UTPlus extensions.
******************************************************************************/


class UTP_Extension extends Info
    hidedropdown
    abstract;


//=============================================================================
// Configuration
//=============================================================================

var config bool bLoadExtension, bShowInServerDetails;


//=============================================================================
// Localization
//=============================================================================

var(UTPlus) const editconst localized string FriendlyName;
var(UTPlus) const editconst localized string Description;

var private localized string ExtensionsGroup;
var private localized string lblShowInServerDetails, descShowInServerDetails;


//=============================================================================
// Variables
//=============================================================================

var MutUTPlus MyMutator;
var bool bReceiveMutate, bNotifyLogin, bNotifyActorSpawn, bOnlyForDedicatedServer, bConfigShowInServerDetails, bRequiresAccessContol;

var private editconst editconstarray noexport array<string> KnownExtensionClassNames/*, KnownExtensionDescriptions*/;
var private editconst editconstarray noexport array<class<UTP_Extension> > KnownExtensionClasses;
var private editconst editconstarray noexport array<byte> FailedToLoadExtension;


//== PostRegister =============================================================
/**
Called during PreBeginPlay() after the extension actor has been registered.
*/
//=============================================================================

function PostRegister(MutUTPlus UTPMutator)
{
  MyMutator = UTPMutator;
}


//== Destroyed ================================================================
/**
Unregister the extension actor.
*/
//=============================================================================

function Destroyed()
{
  if ( MyMutator != None )
    MyMutator.UnregisterExtension(Self);
}


//== PreLogin =================================================================
/**
Called by the UTP_AccessControl when a player attempts to connect to the server. Not called if other PreLogin rejected the connection. 
*/
//=============================================================================

function PreLogin(string Options, string Address, string PlayerID, out string Error, out string FailCode, bool bSpectator);


//== ModifyLogin =================================================================
/**
Called by MutUTPlus right before the player's controller is spawned. 
*/
//=============================================================================

function ModifyLogin(out string Portal, out string Options);


//== PostLogin ================================================================
/**
Called after a player has logged in.
*/
//=============================================================================

function PostLogin(UTP_PlayerInfo LoginPI);


//== PickupQuery ==============================================================
/**
Called when an item is about to be picked up by a player.
*/
//=============================================================================

function PickupQuery(Pawn Other, Pickup Item);


//== Mutate ===================================================================
/**
Called when bReceiveMutate == True and a Mutate console command is received.
*/
//=============================================================================

function Mutate(string MutateString, PlayerController Sender);


//== ActorSpawned =============================================================
/**
Called when bNotifyActorSpawn == True and the UTPlus mutator receives a CheckReplacement call. bReplaced tells, whether another UTP_Extension requested to destroy the actor.
Return True if the actor should be destroyed.
*/
//=============================================================================

function bool ActorSpawned(Actor Other, byte bReplaced);


//== ServerTraveling ==========================================================
/**
Handle map changes.
*/
//=============================================================================

function ServerTraveling(string URL, bool bItems);


//== NickChange ===============================================================
/**
Called when a player changes the nick. For new players RelatedPI.LastPlayerName == "".
*/
//=============================================================================

function NickChange(UTP_PlayerInfo RelatedPI);


//== GetHumanReadableName =====================================================
/**
Returns a human-readable string representation of this actor.
*/
//=============================================================================

simulated function string GetHumanReadableName()
{
  if (FriendlyName != "")
    return FriendlyName;
  
  return super.GetHumanReadableName();
}


//== GetServerDetails =========================================================
/**
Append the extension name to the server details.
*/
//=============================================================================

function GetServerDetails(out GameInfo.ServerResponseLine ServerState)
{
  local int i;
  
  i = ServerState.ServerInfo.Length;
  ServerState.ServerInfo.Length = i + 1;
  ServerState.ServerInfo[i].Key = "UTPExtension";
  ServerState.ServerInfo[i].Value = GetHumanReadableName();
}


//== static ShowExtensionOptions ==============================================
/**
Returns, whether the configuration options of this extension class should be added to the PlayInfo.
*/
//=============================================================================

static function bool ShowExtensionOptions()
{
  local int i;
  
  if ( default.bLoadExtension )
    return true;
  
  for (i = 0; i < class'MutUTPlus'.default.RegisteredExtensionClasses.Length; ++i) {
    if ( class'MutUTPlus'.default.RegisteredExtensionClasses[i] == default.Class )
      return true;
  }
    
  return false;
}


//== static FillPlayInfo ======================================================
/**
Add the configuration options to the PlayInfo. Don't touch this function, override FillExtensionInfo() instead.
*/
//=============================================================================

static function FillPlayInfo(PlayInfo PlayInfo)
{
  if (class'UTP_ServerUtil'.static.ClassIsInPlayInfo(PlayInfo, default.Class))
    return;
  
  Super.FillPlayInfo(PlayInfo);
  
  if ( default.FriendlyName == "" )
    warn("FriendlyName is empty!");
  
  PlayInfo.AddSetting(default.ExtensionsGroup, "bLoadExtension", default.FriendlyName, 250, 0, "Check",, "Xi", True);
  
  if ( ShowExtensionOptions() )
    FillExtensionInfo(PlayInfo);
  
  PlayInfo.PopClass();
}


//== static FillExtensionInfo =================================================
/**
Add the extension's options to the PlayInfo. Should be subclassed.
*/
//=============================================================================

static function FillExtensionInfo(PlayInfo PlayInfo)
{
  if (default.bConfigShowInServerDetails)
    PlayInfo.AddSetting(default.FriendlyName, "bShowInServerDetails", default.lblShowInServerDetails, 250, 0, "Check",, "Xi", True);
}


//== static GetDescriptionText ================================================
/**
Returns a description for this extension.
*/
//=============================================================================

static event string GetDescriptionText(string PropName)
{
  if (PropName == "bLoadExtension")
    return default.Description;
  if (PropName == "bShowInServerDetails")
    return default.descShowInServerDetails;
  
  return Super.GetDescriptionText(PropName);
}


//== static FillExtensionLists ================================================
/**
Returns the list of available UTP_Extension classes as defined in INT files.
*/
//=============================================================================

private static final function FillExtensionLists(optional bool bLoadAllClasses)
{
  local int i;
  local array<string> ExtensionDescriptions;
  
  if (default.KnownExtensionClassNames.Length == 0) {
    default.FailedToLoadExtension.Length = 0;   // clear the failure list
    GetAllIntDesc(string(class'UTP_Extension'), default.KnownExtensionClassNames, ExtensionDescriptions);
    default.KnownExtensionClasses.Length = default.KnownExtensionClassNames.Length;
  }
  default.FailedToLoadExtension.Length = default.KnownExtensionClassNames.Length;
  if (bLoadAllClasses) {
    for (i = 0; i < default.KnownExtensionClassNames.Length; i++) {
      if (default.KnownExtensionClasses[i] == None && default.FailedToLoadExtension[i] == 0) {
        default.KnownExtensionClasses[i] = class<UTP_Extension>(DynamicLoadObject(default.KnownExtensionClassNames[i], class'Class', True));
        if (default.KnownExtensionClasses[i] == None)
          default.FailedToLoadExtension[i] = 1;
      }
    }
  }
  else {
    for (i = 0; i < default.KnownExtensionClasses.Length; i++) {
      if (default.KnownExtensionClasses[i] == None && default.FailedToLoadExtension[i] == 0) {
        default.KnownExtensionClasses[i] = class<UTP_Extension>(FindObject(default.KnownExtensionClassNames[i], class'Class'));
        if (default.KnownExtensionClasses[i] == None)
          default.FailedToLoadExtension[i] = 1;
      }
    }
  }
}


//== static GetExtensionDetails ===============================================
/**
Fills the arrays with the available extension details. bLoadAllClasses forces loading of extension classes.
*/
//=============================================================================

/*static final function GetExtensionDetails(out array<string> ExtensionNames, out array<string> ExtensionDescriptions)
{
  local int i;
  
  // find the extensions in INT files
  FillExtensionLists();
  
  // fill names and descriptions
  ExtensionNames.Length         = default.KnownExtensionClasses.Length;
  ExtensionDescriptions.Length  = default.KnownExtensionClasses.Length;
  for (i = 0; i < default.KnownExtensionClasses.Length; i++) {
    if (default.KnownExtensionClasses[i] != None) {
      // ...from the localized class properties
      ExtensionNames[i]         = default.KnownExtensionClasses[i].default.FriendlyName;
      ExtensionDescriptions[i]  = default.KnownExtensionClasses[i].default.Description;
    }
    else {
      // ...from the INT file object entries (oldskool mutator-style)
      if (!Divide(default.KnownExtensionDescriptions[i], ",", ExtensionNames[i], ExtensionDescriptions[i])) {
        ExtensionNames[i]           = default.KnownExtensionDescriptions[i];
        ExtensionDescriptions[i]    = default.KnownExtensionDescriptions[i];
      }
    }
  }
}*/


//== static GetExtensionList ==================================================
/**
Returns the list of available UTP_Extension classes as defined in INT files.
*/
//=============================================================================

static final function GetExtensionClasses(out array<class<UTP_Extension> > ExtensionClasses, out array<string> ExtensionClassNames, optional bool bLoadAllClasses)
{
  FillExtensionLists(bLoadAllClasses);
  
  ExtensionClasses      = default.KnownExtensionClasses;
  ExtensionClassNames   = default.KnownExtensionClassNames;
}


defaultproperties
{
  ExtensionsGroup = "UTPlus Extensions"
  
  lblShowInServerDetails    = "Show in server details"
  descShowInServerDetails   = "Show this UTPlus extension in the server details."
  
  bLoadExtension            = False
  bOnlyForDedicatedServer   = False
  bShowInServerDetails      = True
  bConfigShowInServerDetails= False
  bRequiresAccessContol     = False
}
