/******************************************************************************
UTPlus
Copyright (c) 2005-2006 by Wormbo

This program is free software; you can redistribute and/or modify 
it under the terms of the Open Unreal Mod License version 1.1.
http://wiki.beyondunreal.com/wiki/OpenUnrealModLicense


Manages the server-side player ID database.
******************************************************************************/


class UTP_PlayerIDs extends UTP_PlayerIDsBase
    transient;


//=============================================================================
// Variables
//=============================================================================

var protected array<GUID> GUIDs; // sync'd with UTP_PlayerIDsBase.IDs array


//== ImportDatabase ===========================================================
/**
Imports a the player ID database from the specified object.
*/
//=============================================================================

function ImportDatabase(UTP_PlayerIDsExport Database)
{
  local int i, j;
  local byte hash;
  
  if ( Database == None ) {
    log("No database object available for loading the player ID database!", 'UTPlusError');
    return;
  }

  log("Reading" @ Database.PlayerList.Length @ "entries (" $ Database.GUIDList.Length @ "GUIDs," @ Database.NameList.Length @ "names) from player ID database...", 'UTPlus');
  
  // pre-allocate memory
  PlayerList.Length = Database.PlayerList.Length + 1;
  Names.Length      = Database.NameList.Length;
  IDs.Length        = Database.GUIDList.Length;
  
  // copy GUIDs
  GUIDs = Database.GUIDList;
  
  // clear hash table
  for (i = 0; i < ArrayCount(HashTable); ++i) {
    HashTable[i].iIDs.Length    = 0;
    HashTable[i].iIPs.Length    = 0;
    HashTable[i].iNames.Length  = 0;
  }
  
  // add names
  for (i = 0; i < Names.Length; ++i) {
    Names[i].PlayerName = Database.NameList[i];
    hash                = HashString(Database.NameList[i]);
    HashTable[hash].iNames[HashTable[hash].iNames.Length] = i;
  }
  
  // add GUIDs
  for (i = 0; i < GUIDs.Length; ++i) {
    IDs[i].ID = IDFromGUID(GUIDs[i]);
    hash      = HashInt(IDs[i].ID);
    HashTable[hash].iIDs[HashTable[hash].iIDs.Length] = i;
  }
  
  // add IPs and fill player list
  for (i = 1; i < PlayerList.Length; ++i) {
    PlayerList[i].iID = Database.PlayerList[i - 1].iID;
    PlayerList[i].iIP = FindIP(Database.PlayerList[i - 1].IP, True);    // also updates the IP hashtable
    PlayerList[i].iNames = Database.PlayerList[i - 1].iNames;
    
    // fill in iPlayer references
    IDs[PlayerList[i].iID].iPlayers[IDs[PlayerList[i].iID].iPlayers.Length] = i;
    IPs[PlayerList[i].iIP].iPlayers[IPs[PlayerList[i].iIP].iPlayers.Length] = i;
    for (j = 0; j < PlayerList[i].iNames.Length; ++j) {
      Names[
        PlayerList[i].iNames[j]
      ].iPlayers[
        Names[
          PlayerList[i].iNames[j]
        ].iPlayers.Length
      ] = i;
    }
  }
}


//== ExportDatabase ===========================================================
/**
Exports the player ID database to the specified object.
*/
//=============================================================================

function ExportDatabase(UTP_PlayerIDsExport Database)
{
  local int i;
  
  if ( Database == None ) {
    log("No database object available for saving the player ID database!", 'UTPlusError');
    return;
  }
  
  log("Storing" @ PlayerList.Length @ "entries (" $ GUIDs.Length @ "GUIDs," @ Names.Length @ "names) in player ID database '" $ Database.Outer $ "'...", 'UTPlus');
  
  // pre-allocate memory
  Database.PlayerList.Length  = PlayerList.Length - 1;
  Database.NameList.Length    = Names.Length;
  
  // copy GUIDs
  Database.GUIDList = GUIDs;
  
  // copy players and IPs
  for (i = 1; i < PlayerList.Length; ++i) {
    Database.PlayerList[i - 1].iID    = PlayerList[i].iID;
    Database.PlayerList[i - 1].IP     = IPs[PlayerList[i].iIP].IP;
    Database.PlayerList[i - 1].iNames = PlayerList[i].iNames;
  }
  
  // copy names
  for (i = 0; i < Names.Length; ++i) {
    Database.NameList[i] = Names[i].PlayerName;
  }
}


//== LogPlayer ================================================================
/**
Adds a new player to the database and returns the player's database ID.
*/
//=============================================================================

function int LogPlayer(GUID PlayerID, int PlayerIP, string PlayerName)
{
  local GUID EmptyGUID;
  
  if ( PlayerID == EmptyGUID || PlayerIP == 0 || PlayerName == "" )
    return -1;
  
  FindGUID(PlayerID, True);
  
  return AddPlayer(IDFromGUID(PlayerID), PlayerIP, PlayerName);
}


//== FindGUID =================================================================
/**
Returns the index of a GUID.
*/
//=============================================================================

function int FindGUID(GUID PlayerID, optional bool bAdd)
{
  local int iID;
  
  iID = FindID(IDFromGUID(PlayerID), bAdd);
  if (iID == GUIDs.Length)
    GUIDs[iID] = PlayerID;
  
  return iID;
}


//== GetPlayerInfo ============================================================
/**
A player's info by UID.
*/
//=============================================================================

function UTP_ServerUtil.TPlayerInfo GetPlayerInfo(int iPlayerUID)
{
  local int i;
  local UTP_ServerUtil.TPlayerInfo PlayerInfo;
  
  if ( iPlayerUID < 0 || iPlayerUID >= PlayerList.Length ) {
    log("UTP_PlayerIDs.GetPlayerInfo - PlayerUID out of bounds:" @ iPlayerUID $ "/" $ PlayerList.Length, 'UTPlusError');
  }
  else {
    PlayerInfo.IP = IPs[PlayerList[iPlayerUID].iIP].IP;
    PlayerInfo.ID = GUIDs[PlayerList[iPlayerUID].iID];
    PlayerInfo.Names.Length = PlayerList[iPlayerUID].iNames.Length;
    for (i = 0; i < PlayerInfo.Names.Length; ++i)
      PlayerInfo.Names[i] = Names[PlayerList[iPlayerUID].iNames[i]].PlayerName;
    class'UTP_ServerUtil'.static.SortS(PlayerInfo.Names, True);
  }
  return PlayerInfo;
}


//== IDFromGUID ===============================================================
/**
Calculates an ID value for a GUID.
*/
//=============================================================================

final function int IDFromGUID(GUID g)
{
  return g.A ^ g.B ^ g.C ^ g.D;
}
