/******************************************************************************
UTPlus
Copyright (c) 2005-2006 by Wormbo

This program is free software; you can redistribute and/or modify 
it under the terms of the Open Unreal Mod License version 1.1.
http://wiki.beyondunreal.com/wiki/OpenUnrealModLicense


Abstract class for utility functions used by the UTPlus server-side package.
******************************************************************************/


class UTP_ServerUtil extends UTP_Util
    abstract;


//=============================================================================
// Constants
//=============================================================================

const VALID_NICKNAME_CHARS = "!#$%&'()*+,-./0123456789:;<=>@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~";
const VALID_URL_CHARS = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz-_.!~*'()";
const HEX_DIGITS = "0123456789ABCDEF";


//=============================================================================
// Structs
//=============================================================================

struct TPlayerInfo {
  var GUID          ID;
  var int           IP;
  var array<string> Names;
};


//== static GetVersionString ==================================================
/**
Returns the UTPlus version and build.
*/
//=============================================================================

static function string GetVersionString(optional bool bNoBuild)
{
  if ( bNoBuild )
    return class'MutUTPlus'.default.FriendlyName;
  else
    return class'MutUTPlus'.default.FriendlyName @ "build" @ class'MutUTPlus'.default.Build;
}


//== static GetCopyrightString ================================================
/**
Returns the UTPlus copyright notice.
*/
//=============================================================================

static function string GetCopyrightString()
{
  return class'MutUTPlus'.default.Copyright;
}


//== static GUIDToID ==========================================================
/**
Reads a GUID and returns a PlayerID string.
*/
//=============================================================================

static final function string GUIDToID(GUID ID, optional bool bLowerCase, optional bool bAddSeparator, optional bool bMaskID)
{
  local string IDString;
  
  if ( bAddSeparator )
    IDString = IntToHex(ID.A, bLowerCase) $ "-" $ IntToHex(ID.B, bLowerCase) $ "-" $ IntToHex(ID.C, bLowerCase) $ "-";
  else
    IDString = IntToHex(ID.A, bLowerCase) $ IntToHex(ID.B, bLowerCase) $ IntToHex(ID.C, bLowerCase);
  if ( bMaskID )
    return IDString $ "********";
  else
    return IDString $ IntToHex(ID.D, bLowerCase);
}


//== static IntToHex ==========================================================
/**
Converts an integer to hexadecimal string.
*/
//=============================================================================

static final function string IntToHex(int i, optional bool bLowerCase, optional bool bNoLeadingZeroes)
{
  local string Result;
  
  while (i != 0) {
    if ( (i & 0xF) < 10 )
      Result = (i & 0xF) $ Result;
    else
      Result = Chr(55 + (i & 0xF)) $ Result;
    i = i >>> 4;
  }
  if ( bLowerCase )
    Result = Locs(Result);
  
  if ( bNoLeadingZeroes )
    return Result;
  return Right("00000000" $ Result, 8);
}


//== static CompareGUIDs ======================================================
/**
Compares two GUIDs and returns whether the first comes before the second.
*/
//=============================================================================

static final function bool CompareGUIDs(GUID A, GUID B)
{
  if ( A.A < B.A )
    return true;
  if ( A.A > B.A )
    return false;
  
  // It's highly unlikely two IDs start with 8 common digits without being
  // identical, so let's try the faster (native) operator == (struct, struct)
  if ( A == B )
    return false;
  
  if ( A.B < B.B )
    return true;
  if ( A.B > B.B )
    return false;
  
  if ( A.C < B.C )
    return true;
  if ( A.C > B.C )
    return false;
  
  return A.D < B.D;
}


//== static HasParameter ======================================================
/**
Looks for a parameter in a string.
*/
//=============================================================================

static final function bool HasParameter(string AllParameters, string Parameter)
{
  if ( Left(AllParameters, Len(Parameter)) ~= Parameter )
    return true;
  if ( Right(AllParameters, Len(Parameter)) ~= Parameter )
    return true;
  if ( InStr(AllParameters, "" @ Parameter @ "") != -1 )
    return true;
  
  return false;
}


//== static GetPackageName ====================================================
/**
Extracts the package name from an object string.
*/
//=============================================================================

static final function string GetPackageName(coerce string S)
{
  local int i;
  
  i = InStr(S, ".");
  if ( i != -1 )
    S = Left(S, i);
  
  return S;
}


//== static ClassIsInPlayInfo ====================================================
/**
Extracts the package name from an object string.
*/
//=============================================================================

static final function bool ClassIsInPlayInfo(PlayInfo PlayInfo, class<Info> InfoClass)
{
  local int i;
  
  if ( PlayInfo != None && InfoClass != None ) {
    for (i = 0; i < PlayInfo.InfoClasses.Length; ++i) {
      if ( PlayInfo.InfoClasses[i] == InfoClass )
        return true;
    }
  }
  return false;
}


//== static StripNick =========================================================
/**
Strips color codes and illegal characters from a player name.
*/
//=============================================================================

static final function string StripNick(coerce string InputString)
{
  local string OutputString, LeftChar;
  
  InputString = StripCodes(Left(InputString, 20));
  while(InputString != "") {
    LeftChar = Left(InputString, 1);
    if (InStr(VALID_NICKNAME_CHARS, LeftChar) != -1)
      OutputString $= LeftChar;
    else
      OutputString $= "_";
    InputString = Mid(InputString, 1);
  }
  return OutputString;
}


//== static URLEncode =========================================================
/**
Encodes a string so it can be used in a HTTP request URL.
*/
//=============================================================================

static final function string URLEncode(coerce string InputString)
{
  local string OutputString, LeftChar;
  
  while(InputString != "") {
    LeftChar = Left(InputString, 1);
    if (InStr(VALID_URL_CHARS, LeftChar) != -1)
      OutputString $= LeftChar;
    else
      OutputString $= "%" $ Mid(HEX_DIGITS, Asc(LeftChar) >>> 4, 1) $ Mid(HEX_DIGITS, Asc(LeftChar) & 0xF, 1);
    InputString = Mid(InputString, 1);
  }
  return OutputString;
}


//=============================================================================
// Default properties
//=============================================================================

defaultproperties
{
}
